"""
PVPositioner that computes ``done`` as a soft signal.

.. autosummary::

   ~PVPositionerSoftDone
   ~PVPositionerSoftDoneWithStop
"""

from ..tests import short_delay_for_EPICS_IOC_database_processing
from ophyd import Component
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import FormattedComponent
from ophyd import PVPositioner
from ophyd import Signal
from ophyd.signal import EpicsSignalBase
import logging


logger = logging.getLogger(__name__)


class PVPositionerSoftDone(PVPositioner):
    """
    PVPositioner that computes ``done`` as a soft signal.

    PARAMETERS

    prefix : str, optional
        The device prefix used for all sub-positioners.
        This is optional as it
        may be desirable to specify full PV names for PVPositioners.
    readback_pv : str, optional
        PV prefix of the readback signal.
        Disregarded if readback attribute is created.
    setpoint_pv : str, optional
        PV prefix of the setpoint signal.
        Disregarded if setpoint attribute is created.
    tolerance : float, optional
        Motion tolerance. The motion is considered *done* when::

            abs(readback-setpoint) <= tolerance

        Defaults to ``10^(-1*precision)``,
        where ``precision = setpoint.precision``.
    update_target : bool
        ``True`` when this object update the ``target`` Component directly.
        Use ``False`` if the ``target`` Component will be updated externally,
        such as by the controller when ``target`` is an ``EpicsSignal``.
        Defaults to ``True``.
    kwargs :
        Passed to `ophyd.PVPositioner`

    ATTRIBUTES

    setpoint : Signal
        The setpoint (request) signal
    readback : Signal or None
        The readback PV (e.g., encoder position PV)
    actuate : Signal or None
        The actuation PV to set when movement is requested
    actuate_value : any, optional
        The actuation value, sent to the actuate signal when motion is
        requested
    stop_signal : Signal or None
        The stop PV to set when motion should be stopped
    stop_value : any, optional
        The value sent to stop_signal when a stop is requested
    target : Signal
        The target value of a move request.

        Override (in subclass) with `EpicsSignal` to connect with a PV.

        In some controllers (such as temperature controllers), the setpoint
        may be changed incrementally towards this target value (such as a
        ramp or controlled trajectory).  In such cases, the ``target`` will
        be final value while ``setpoint`` will be the current desired position.

        Otherwise, both ``setpoint`` and ``target`` will be set to the same value.

    (new in apstools 1.5.3)
    """

    # positioner
    readback = FormattedComponent(
        EpicsSignalRO, "{prefix}{_readback_pv}", kind="hinted", auto_monitor=True
    )
    setpoint = FormattedComponent(
        EpicsSignal, "{prefix}{_setpoint_pv}", kind="normal", put_complete=True
    )
    done = Component(Signal, value=True, kind="config")
    done_value = True

    tolerance = Component(Signal, value=-1, kind="config")
    report_dmov_changes = Component(Signal, value=False, kind="omitted")

    target = Component(Signal, value="None", kind="config")

    _rb_count = 0
    _sp_count = 0

    @property
    def precision(self):
        return self.setpoint.precision

    @property
    def actual_tolerance(self):
        return (
            self.tolerance.get()
            if self.tolerance.get() >= 0
            else 10 ** (-1 * self.precision)
        )

    def cb_readback(self, *args, **kwargs):
        """
        Called when readback changes (EPICS CA monitor event).

        Computes if the positioner is done moving::

            done = |readback - setpoint| <= tolerance
        """
        self._rb_count += 1
        if "value" in kwargs:
            diff = kwargs["value"] - self.setpoint.get()
        else:
            diff = self.readback.get(use_monitor=False) - self.setpoint.get()
        dmov = abs(diff) <= self.actual_tolerance
        if self.report_dmov_changes.get() and dmov != self.done.get():
            logger.debug("%s reached: %s", self.name, dmov)

        v = self.done_value
        self.done.put({True: v, False: not v}[dmov])
        logger.debug("cb_readback: done=%s, position=%s", self.done.get(), self.position)

    def cb_setpoint(self, *args, **kwargs):
        """
        Called when setpoint changes (EPICS CA monitor event).

        When the setpoint is changed, force done=False.  For any move, done
        **must** transition to ``!= done_value``, then back to ``done_value``.

        Without this response, a small move (within tolerance) will not return.
        Next update of readback will compute ``self.done``.
        """
        self._sp_count += 1
        self.done.put(not self.done_value)
        logger.debug("cb_setpoint: done=%s, setpoint=%s", self.done.get(), self.setpoint.get())

    def __init__(
        self,
        prefix="",
        *,
        readback_pv="",
        setpoint_pv="",
        tolerance=None,
        update_target=True,
        **kwargs,
    ):

        if setpoint_pv == readback_pv:
            raise ValueError(
                f"readback_pv ({readback_pv})"
                f" and setpoint_pv ({setpoint_pv})"
                " must have different values"
            )
        self._setpoint_pv = setpoint_pv
        self._readback_pv = readback_pv

        super().__init__(prefix=prefix, **kwargs)

        # Make the default alias for the readback the name of the
        # positioner itself as in EpicsMotor.
        self.readback.name = self.name
        self.update_target = update_target

        self.readback.subscribe(self.cb_readback)
        self.setpoint.subscribe(self.cb_setpoint)
        if tolerance:
            self.tolerance.put(tolerance)

    def _setup_move(self, position):
        """Move and do not wait until motion is complete (asynchronous)"""
        self.log.debug("%s.setpoint = %s", self.name, position)
        if self.update_target:
            kwargs = {}
            if issubclass(self.target.__class__, EpicsSignalBase):
                kwargs["wait"] = True  # Signal.put() warns if kwargs are given
            self.target.put(position, **kwargs)
        self.setpoint.put(position, wait=True)
        if self.actuate is not None:
            self.log.debug("%s.actuate = %s", self.name, self.actuate_value)
            self.actuate.put(self.actuate_value, wait=False)
        # This is needed because in a special case the setpoint.put does not
        # run the "sub_value" subscriptions.
        self.cb_setpoint()
        self.cb_readback()  # This is needed to force the first check.


class PVPositionerSoftDoneWithStop(PVPositionerSoftDone):
    """
    PVPositionerSoftDone with stop() and inposition.

    The :meth:`stop()` method sets the setpoint to the immediate readback
    value (only when ``inposition`` is ``True``).  This stops the
    positioner at the current position.
    """

    @property
    def inposition(self):
        """
        Report (boolean) if positioner is done.
        """
        return self.done.get() == self.done_value

    def stop(self, *, success=False):
        """
        Hold the current readback when stop() is called and not :meth:`inposition`.
        """
        if not self.inposition:
            self.setpoint.put(self.position)
            short_delay_for_EPICS_IOC_database_processing()
            self.cb_readback()  # re-evaluate soft done Signal

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Support for APS hardware abstractions (both physical and virtual).
"""

# must come first to avoid circular imports
from .positioner_soft_done import PVPositionerSoftDone
from .positioner_soft_done import PVPositionerSoftDoneWithStop

# other imports
from .aps_bss_user import ApsBssUserInfoDevice

from .aps_cycle import ApsCycleDM

from .aps_machine import ApsMachineParametersDevice

from .aps_undulator import ApsUndulator
from .aps_undulator import ApsUndulatorDual

from .area_detector_support import AD_EpicsFileNameMixin
from .area_detector_support import AD_FrameType_schemes
from .area_detector_support import AD_plugin_primed
from .area_detector_support import AD_prime_plugin
from .area_detector_support import AD_prime_plugin2
from .area_detector_support import AD_full_file_name_local
from .area_detector_support import AD_EpicsFileNameHDF5Plugin
from .area_detector_support import AD_EpicsFileNameJPEGPlugin
from .area_detector_support import AD_EpicsFileNameTIFFPlugin
from .area_detector_support import AD_EpicsHdf5FileName
from .area_detector_support import AD_EpicsHDF5IterativeWriter
from .area_detector_support import AD_EpicsJPEGFileName
from .area_detector_support import AD_EpicsJPEGIterativeWriter
from .area_detector_support import AD_EpicsTIFFFileName
from .area_detector_support import AD_EpicsTIFFIterativeWriter
from .area_detector_support import CamMixin_V34
from .area_detector_support import CamMixin_V3_1_1
from .area_detector_support import SingleTrigger_V34


from .axis_tuner import AxisTunerException
from .axis_tuner import AxisTunerMixin

from .description_mixin import EpicsDescriptionMixin

from .dict_device_support import dict_device_factory
from .dict_device_support import make_dict_device

from .epics_scan_id_signal import EpicsScanIdSignal

from .eurotherm_2216e import Eurotherm2216e

from .flyer_motor_scaler import FlyerBase
from .flyer_motor_scaler import ActionsFlyerBase
from .flyer_motor_scaler import ScalerMotorFlyer
from .flyer_motor_scaler import SignalValueStack
from .flyer_motor_scaler import _SMFlyer_Step_1
from .flyer_motor_scaler import _SMFlyer_Step_2
from .flyer_motor_scaler import _SMFlyer_Step_3

from .kohzu_monochromator import KohzuSeqCtl_Monochromator

from .lakeshore_controllers import LakeShore336Device
from .lakeshore_controllers import LakeShore340Device

from .linkam_controllers import Linkam_CI94_Device
from .linkam_controllers import Linkam_T96_Device

from .mixin_base import DeviceMixinBase

from .motor_mixins import EpicsMotorDialMixin
from .motor_mixins import EpicsMotorEnableMixin
from .motor_mixins import EpicsMotorLimitsMixin
from .motor_mixins import EpicsMotorRawMixin
from .motor_mixins import EpicsMotorResolutionMixin
from .motor_mixins import EpicsMotorServoMixin

from .ptc10_controller import PTC10AioChannel
from .ptc10_controller import PTC10RtdChannel
from .ptc10_controller import PTC10TcChannel
from .ptc10_controller import PTC10PositionerMixin

from .scaler_support import SCALER_AUTOCOUNT_MODE
from .scaler_support import use_EPICS_scaler_channels

from .shutters import ApsPssShutter
from .shutters import ApsPssShutterWithStatus
from .shutters import EpicsMotorShutter
from .shutters import EpicsOnOffShutter
from .shutters import OneSignalShutter
from .shutters import ShutterBase
from .shutters import SimulatedApsPssShutterWithStatus

from .srs570_preamplifier import SRS570_PreAmplifier

from .struck3820 import Struck3820

from .synth_pseudo_voigt import SynPseudoVoigt

from .tracking_signal import TrackingSignal

from .xia_pf4 import DualPf4FilterBox
from .xia_pf4 import Pf4FilterBank
from .xia_pf4 import Pf4FilterCommon
from .xia_pf4 import Pf4FilterDual
from .xia_pf4 import Pf4FilterSingle
from .xia_pf4 import Pf4FilterTriple

from .xia_slit import XiaSlit2D

# synApps

# ## _common
from ..synApps import EpicsRecordDeviceCommonAll
from ..synApps import EpicsRecordInputFields
from ..synApps import EpicsRecordOutputFields
from ..synApps import EpicsRecordFloatFields
from ..synApps import EpicsSynAppsRecordEnableMixin

# ## asyn
from ..synApps import AsynRecord

# ## busy
from ..synApps import BusyRecord

# ## calcout
from ..synApps import CalcoutRecord
from ..synApps import CalcoutRecordChannel
from ..synApps import setup_gaussian_calcout
from ..synApps import setup_incrementer_calcout
from ..synApps import setup_lorentzian_calcout
from ..synApps import UserCalcoutDevice
from ..synApps import UserCalcoutN

# ## epid
from ..synApps import EpidRecord

# ## iocstats
from ..synApps import IocStatsDevice

# ## save_data
from ..synApps import SaveData

# ## scalcout
from ..synApps import UserScalcoutDevice
from ..synApps import UserScalcoutN
from ..synApps import ScalcoutRecord
from ..synApps import ScalcoutRecordNumberChannel
from ..synApps import ScalcoutRecordStringChannel

# ## sscan
from ..synApps import SscanRecord
from ..synApps import SscanDevice

# sseq
from ..synApps import EditStringSequence
from ..synApps import SseqRecord
from ..synApps import UserStringSequenceDevice
from ..synApps import UserStringSequenceN

# ## sub
from ..synApps import SubRecord
from ..synApps import SubRecordChannel
from ..synApps import UserAverageN
from ..synApps import UserAverageDevice

# ## swait
from ..synApps import SwaitRecord
from ..synApps import SwaitRecordChannel
from ..synApps import UserCalcN
from ..synApps import UserCalcsDevice
from ..synApps import setup_random_number_swait
from ..synApps import setup_gaussian_swait
from ..synApps import setup_lorentzian_swait
from ..synApps import setup_incrementer_swait

# ## transform
from ..synApps import TransformRecord
from ..synApps import UserTransformN
from ..synApps import UserTransformsDevice

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Spreadsheet Support
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~ExcelDatabaseFileBase
   ~ExcelDatabaseFileGeneric
   ~ExcelReadError
"""

import math
import openpyxl
import openpyxl.utils.exceptions
import os

from collections import OrderedDict

from . import to_unicode_or_bust


class ExcelReadError(openpyxl.utils.exceptions.InvalidFileException):
    """
    Exception when reading Excel spreadsheet.

    .. index:: apstools Exception; ExcelReadError
    """


class ExcelDatabaseFileBase(object):
    """
    base class: read-only support for Excel files, treat them like databases

    .. index:: apstools Utility; ExcelDatabaseFileBase

    Use this class when creating new, specific spreadsheet support.

    EXAMPLE

    Show how to read an Excel file where one of the columns
    contains a unique key.  This allows for random access to
    each row of data by use of the *key*.

    ::

        class ExhibitorsDB(ExcelDatabaseFileBase):
            '''
            content for exhibitors from the Excel file
            '''
            EXCEL_FILE = os.path.join("resources", "exhibitors.xlsx")
            LABELS_ROW = 2

            def handle_single_entry(self, entry):
                '''any special handling for a row from the Excel file'''
                pass

            def handleExcelRowEntry(self, entry):
                '''identify unique key (row of the Excel file)'''
                key = entry["Name"]
                self.db[key] = entry

    """

    EXCEL_FILE = None  # subclass MUST define
    # EXCEL_FILE = os.path.join("abstracts", "index of abstracts.xlsx")
    LABELS_ROW = 3  # labels are on line LABELS_ROW+1 in the Excel file

    def __init__(self, ignore_extra=True):
        self.db = OrderedDict()
        self.data_labels = None
        if self.EXCEL_FILE is None:
            raise ValueError("subclass must define EXCEL_FILE")
        self.fname = os.path.join(os.getcwd(), self.EXCEL_FILE)

        self.sheet_name = 0

        self.parse(ignore_extra=ignore_extra)

    def handle_single_entry(self, entry):  # subclass MUST override
        # fmt: off
        raise NotImplementedError(
            "subclass must override handle_single_entry() method"
        )
        # fmt: on

    def handleExcelRowEntry(self, entry):  # subclass MUST override
        # fmt: off
        raise NotImplementedError(
            "subclass must override handleExcelRowEntry() method"
        )
        # fmt: on

    def parse(
        self, labels_row_num=None, data_start_row_num=None, ignore_extra=True,
    ):
        labels_row_num = labels_row_num or self.LABELS_ROW
        try:
            wb = openpyxl.load_workbook(self.fname)
            ws = wb.worksheets[self.sheet_name]
            if ignore_extra:
                # ignore data outside of table in spreadsheet file
                data = list(ws.rows)[labels_row_num:]
                self.data_labels = []
                for c in data[0]:
                    if c.value is None:
                        break
                    self.data_labels.append(c.value)
                rows = []
                for r in data[1:]:
                    if r[0].value is None:
                        break
                    rows.append(r[: len(self.data_labels)])
            else:
                # use the whole sheet
                rows = list(ws.rows)
                # create the column titles
                # fmt: off
                self.data_labels = [
                    f"Column_{i+1}" for i in range(len(rows[0]))
                ]
                # fmt: on
        except openpyxl.utils.exceptions.InvalidFileException as exc:
            raise ExcelReadError(exc)
        for row in rows:
            entry = OrderedDict()
            for _col, label in enumerate(self.data_labels):
                entry[label] = row[_col].value
                self.handle_single_entry(entry)
            self.handleExcelRowEntry(entry)

    def _getExcelColumnValue(self, row_data, col):
        v = row_data[col]
        if self._isExcel_nan(v):
            v = None
        else:
            v = to_unicode_or_bust(v)
            if isinstance(v, str):
                v = v.strip()
        return v

    def _isExcel_nan(self, value):
        if not isinstance(value, float):
            return False
        return math.isnan(value)


class ExcelDatabaseFileGeneric(ExcelDatabaseFileBase):
    """
    Generic (read-only) handling of Excel spreadsheet-as-database

    .. index:: apstools Utility; ExcelDatabaseFileGeneric
    .. index:: Excel scan, scan; Excel

    .. note:: This is the class to use when reading Excel spreadsheets.

    In the spreadsheet, the first sheet should contain the table to be
    used. By default (see keyword parameter ``labels_row``), the table
    should start in cell A4.  The column labels are given in row 4.  A
    blank column should appear to the right of the table (see keyword
    parameter ``ignore_extra``). The column labels will describe the
    action and its parameters.  Additional columns may be added for
    metadata or other purposes.

    The rows below the column labels should contain actions and
    parameters for those actions, one action per row.

    To make a comment, place a ``#`` in the action column.  A comment
    should be ignored by the bluesky plan that reads this table. The
    table will end with a row of empty cells.

    While it's a good idea to put the ``action`` column first, that is
    not necessary. It is not even necessary to name the column
    ``action``. You can re-arrange the order of the columns and change
    their names **as long as** the column names match what text strings
    your Python code expects to find.

    A future upgrade [#]_ will allow the table boundaries to be named by
    Excel when using Excel's ``Format as Table`` [#]_ feature. For now,
    leave a blank row and column at the bottom and right edges of the
    table.

    .. [#] https://github.com/BCDA-APS/apstools/issues/122
    .. [#] Excel's ``Format as Table``:
        https://support.office.com/en-us/article/Format-an-Excel-table-6789619F-C889-495C-99C2-2F971C0E2370

    PARAMETERS

    filename
        *str* :
        name (absolute or relative) of Excel spreadsheet file
    labels_row
        *int* :
        Row (zero-based numbering) of Excel file with column labels,
        default: ``3`` (Excel row 4)
    ignore_extra
        *bool* :
        When ``True``, ignore any cells outside of the table, default:
        ``True``.

        Note that when ``True``, a row of cells *within* the table will
        be recognized as the end of the table, even if there are
        actions in following rows.  To force an empty row, use
        a comment symbol ``#`` (actually, any non-empty content will work).

        When ``False``, cells with other information (in Sheet 1) will
        be made available, sometimes with unpredictable results.

    EXAMPLE

    See section :ref:`example_run_command_file` for more examples.

    (See also :ref:`example screen shot
    <excel_plan_spreadsheet_screen>`.) Table (on Sheet 1) begins on row
    4 in first column::

        1  |  some text here, maybe a title
        2  |  (could have content here)
        3  |  (or even more content here)
        4  |  action  | sx   | sy   | sample     | comments          |  | <-- leave empty column
        5  |  close   |      |                   | close the shutter |  |
        6  |  image   | 0    | 0    | dark       | dark image        |  |
        7  |  open    |      |      |            | open the shutter  |  |
        8  |  image   | 0    | 0    | flat       | flat field image  |  |
        9  |  image   | 5.1  | -3.2 | 4140 steel | heat 9172634      |  |
        10 |  scan    | 5.1  | -3.2 | 4140 steel | heat 9172634      |  |
        11 |  scan    | 0    | 0    | blank      |                   |  |
        12 |
        13 |  ^^^ leave empty row ^^^
        14 | (could have content here)



    Example python code to read this spreadsheet::

        from apstools.utils import ExcelDatabaseFileGeneric, cleanupText

        def myExcelPlan(xl_file, md={}):
            excel_file = os.path.abspath(xl_file)
            xl = ExcelDatabaseFileGeneric(excel_file)
            for i, row in xl.db.values():
                # prepare the metadata
                _md = {cleanupText(k): v for k, v in row.items()}
                _md["xl_file"] = xl_file
                _md["excel_row_number"] = i+1
                _md.update(md) # overlay with user-supplied metadata

                # determine what action to take
                action = row["action"].lower()
                if action == "open":
                    yield from bps.mv(shutter, "open")
                elif action == "close":
                    yield from bps.mv(shutter, "close")
                elif action == "image":
                    # your code to take an image, given **row as parameters
                    yield from my_image(**row, md=_md)
                elif action == "scan":
                    # your code to make a scan, given **row as parameters
                    yield from my_scan(**row, md=_md)
                else:
                    print(f"no handling for row {i+1}: action={action}")

        # execute this plan through the RunEngine
        RE(myExcelPlan("spreadsheet.xlsx", md=dict(purpose="apstools demo"))

    """

    def __init__(self, filename, labels_row=3, ignore_extra=True):
        self._index_ = 0
        self.EXCEL_FILE = self.EXCEL_FILE or filename
        self.LABELS_ROW = labels_row
        ExcelDatabaseFileBase.__init__(self, ignore_extra=ignore_extra)

    def handle_single_entry(self, entry):
        pass

    def handleExcelRowEntry(self, entry):
        """use row number as the unique key"""
        key = str(self._index_)
        self.db[key] = entry
        self._index_ += 1

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from .catalog import copy_filtered_catalog
from .catalog import findCatalogsInNamespace
from .catalog import getCatalog
from .catalog import getDatabase
from .catalog import getDefaultCatalog
from .catalog import getDefaultDatabase
from .catalog import getStreamValues

from .device_info import listdevice

from .email import EmailNotifications

from .image_analysis import analyze_1D
from .image_analysis import analyze_2D

from .list_plans import listplans

from .list_runs import getRunData
from .list_runs import getRunDataValue
from .list_runs import listRunKeys
from .list_runs import listruns
from .list_runs import ListRuns
from .list_runs import summarize_runs


from .log_utils import file_log_handler
from .log_utils import get_log_path
from .log_utils import setup_IPython_console_logging
from .log_utils import stream_log_handler

from .memory import rss_mem

from .misc import cleanupText
from .misc import connect_pvlist
from .misc import dictionary_table
from .misc import full_dotted_name
from .misc import itemizer
from .misc import listobjects
from .misc import pairwise
from .misc import print_RE_md
from .misc import redefine_motor_position
from .misc import replay
from .misc import run_in_thread
from .misc import safe_ophyd_name
from .misc import split_quoted_line
from .misc import text_encode
from .misc import to_unicode_or_bust
from .misc import trim_string_for_EPICS
from .misc import unix

from .override_parameters import OverrideParameters

from .plot import select_live_plot
from .plot import select_mpl_figure
from .plot import trim_plot_by_name
from .plot import trim_plot_lines

from .profile_support import getDefaultNamespace
from .profile_support import ipython_profile_name
from .profile_support import ipython_shell_namespace

from .pvregistry import findbyname
from .pvregistry import findbypv

from .query import db_query

from .slit_core import SlitGeometry

from .spreadsheet import ExcelDatabaseFileBase
from .spreadsheet import ExcelDatabaseFileGeneric
from .spreadsheet import ExcelReadError

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from ophyd import EpicsSignal
import logging

logger = logging.getLogger(__name__)


class EpicsScanIdSignal(EpicsSignal):
    """
    Use an EPICS PV as the source of the RunEngine's ``scan_id``.

    Uses a writable EPICS integer PV (such as longout record).

    EXAMPLE::

        scan_id = EpicsScanIdDevice("ioc:scan_id:longout", name="scan_id")
        # ...
        RE = bluesky.RunEngine({}, scan_id_source=scan_id.cb_scan_id_source)

    (new in release 1.6.3)
    """

    def cb_scan_id_source(self, *args, **kwargs):
        """
        Callback function for RunEngine.  Returns *next* scan_id to be used.

        * Get current scan_id from PV.
        * Apply lower limit of zero.
        * Increment.
        * Set PV with new value.
        * Return new value.
        """
        new_scan_id = max(self.get(), 0) + 1
        self.put(new_scan_id)
        return new_scan_id

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



import pytest

from ..db_2slit import Optics2Slit1D
from ..db_2slit import Optics2Slit2D_HV
from ..db_2slit import Optics2Slit2D_InbOutBotTop
from ...utils import SlitGeometry
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing

# TODO: test "sync" signal


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [Optics2Slit1D, f"{IOC}Slit1H", False, "read_attrs", 12],
        [Optics2Slit1D, f"{IOC}Slit1H", False, "configuration_attrs", 16],
        [Optics2Slit1D, f"{IOC}Slit1H", True, "read()", 8],
        [Optics2Slit1D, f"{IOC}Slit1H", True, "summary()", 63],

        [Optics2Slit2D_HV, f"{IOC}Slit1", False, "read_attrs", 26],
        [Optics2Slit2D_HV, f"{IOC}Slit1", False, "configuration_attrs", 34],
        [Optics2Slit2D_HV, f"{IOC}Slit1", True, "read()", 16],
        [Optics2Slit2D_HV, f"{IOC}Slit1", True, "summary()", 114],

        [Optics2Slit2D_InbOutBotTop, f"{IOC}Slit1", False, "read_attrs", 24],
        [Optics2Slit2D_InbOutBotTop, f"{IOC}Slit1", False, "configuration_attrs", 32],
        [Optics2Slit2D_InbOutBotTop, f"{IOC}Slit1", True, "read()", 16],
        [Optics2Slit2D_InbOutBotTop, f"{IOC}Slit1", True, "summary()", 112],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


def test_2slit1D():
    axis = Optics2Slit1D(f"{IOC}Slit1H", name="axis")
    assert axis is not None

    axis.wait_for_connection()
    assert axis.connected

    cns = "xn xp size center sync".split()
    assert axis.component_names == tuple(cns)


def test_2slit2D_HV():
    slit1 = Optics2Slit2D_HV(f"{IOC}Slit1", name="slit1")
    assert slit1 is not None

    slit1.wait_for_connection()
    assert slit1.connected
    assert slit1.component_names == ("h", "v")

    slit1.h.size.move(0.1)
    slit1.v.size.move(0.2)
    slit1.h.center.move(0.)
    slit1.v.center.move(0.)
    short_delay_for_EPICS_IOC_database_processing()
    assert round(slit1.h.size.position, 4) == 0.1
    assert round(slit1.v.size.position, 4) == 0.2
    assert round(slit1.h.center.position, 4) == 0.
    assert round(slit1.v.center.position, 4) == 0.

    g = slit1.geometry
    assert isinstance(g, SlitGeometry)
    assert g == SlitGeometry(0.1, 0.2, 0, 0)
    assert round(slit1.h.xn.position, 4) == -0.05
    assert round(slit1.h.xp.position, 4) == 0.05
    assert round(slit1.v.xn.position, 4) == -0.1
    assert round(slit1.v.xp.position, 4) == 0.1


def test_2slit2D_InbOutBotTop():
    slit1 = Optics2Slit2D_InbOutBotTop(f"{IOC}Slit1", name="slit1")
    assert slit1 is not None

    slit1.wait_for_connection()
    assert slit1.connected

    cns = """
    inb out bot top
    hsize hcenter
    vsize vcenter
    hsync vsync
    """.split()
    assert slit1.component_names != tuple(cns)

    slit1.hsize.move(0.3)
    slit1.vsize.move(0.4)
    slit1.hcenter.move(0.)
    slit1.vcenter.move(0.)
    short_delay_for_EPICS_IOC_database_processing()
    assert round(slit1.hsize.position, 4) == 0.3
    assert round(slit1.vsize.position, 4) == 0.4
    assert round(slit1.hcenter.position, 4) == 0.
    assert round(slit1.vcenter.position, 4) == 0.

    g = slit1.geometry
    assert isinstance(g, SlitGeometry)
    assert g == SlitGeometry(0.3, 0.4, 0, 0)
    assert round(slit1.inb.position, 4) == -0.15
    assert round(slit1.out.position, 4) == 0.15
    assert round(slit1.bot.position, 4) == -0.2
    assert round(slit1.top.position, 4) == 0.2


@pytest.mark.parametrize(
    "klass, prefix",
    [
        [Optics2Slit2D_HV, f"{IOC}Slit1"],
        [Optics2Slit2D_InbOutBotTop, f"{IOC}Slit1"],
    ]
)
def test_geometry_property(klass, prefix):
    slit1 = klass(prefix, name="slit1")
    assert slit1 is not None

    slit1.wait_for_connection()
    assert slit1.connected

    h, v, x, y = (0.3, 0.1, .5, -.4)
    slit1.geometry = (h, v, x, y)
    assert slit1.geometry == SlitGeometry(h, v, x, y)



"""
Lakeshore temperature controllers
+++++++++++++++++++++++++++++++++

.. autosummary::

   ~LakeShore336Device
   ~LakeShore340Device

.. not yet implemented
   ~LakeShore330Device
   ~LakeShore331Device
   ~LakeShore335Device
   ~LakeShoreDR93CADevice
"""

# TODO: refactor
#    temperature = Component(PVPositionerSoftDoneWithStop, ...)

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import EpicsSignalWithRBV
from ophyd import FormattedComponent
from ophyd import Signal

from . import PVPositionerSoftDoneWithStop
from ..synApps import AsynRecord

SECOND = 1
MINUTE = 60 * SECOND
HOUR = 60 * MINUTE


class LakeShore336_LoopControl(PVPositionerSoftDoneWithStop):
    """
    LakeShore 336 temperature controller -- with heater control.

    The LakeShore 336 accepts up to two heaters.
    """

    # position
    # readback is provided by PVPositionerSoftDoneWithStop

    setpoint = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}OUT{loop_number}:SP", put_complete=True
    )

    # configuration
    units = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}IN{loop_number}:Units", kind="config"
    )

    heater = FormattedComponent(
        EpicsSignalRO,
        "{prefix}HTR{loop_number}",
        auto_monitor=True,
        kind="normal"
    )
    heater_range = FormattedComponent(
        EpicsSignalWithRBV,
        "{prefix}HTR{loop_number}:Range",
        kind="config",
        auto_monitor=True,
        string=True,
    )

    pid_P = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}P{loop_number}", kind="config"
    )
    pid_I = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}I{loop_number}", kind="config"
    )
    pid_D = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}D{loop_number}", kind="config"
    )

    ramp_rate = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}RampR{loop_number}", kind="config"
    )
    ramp_on = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}OnRamp{loop_number}", kind="config"
    )

    loop_name = FormattedComponent(
        EpicsSignalRO, "{prefix}IN{loop_number}:Name_RBV", kind="config"
    )
    control = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}OUT{loop_number}:Cntrl", kind="config"
    )
    manual = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}OUT{loop_number}:MOUT", kind="config"
    )
    mode = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}OUT{loop_number}:Mode", kind="config"
    )

    def __init__(self, *args, loop_number=None, timeout=10 * HOUR, **kwargs):
        self.loop_number = loop_number
        super().__init__(
            *args,
            timeout=timeout,
            tolerance=0.1,
            readback_pv=f"IN{loop_number}",
            **kwargs
        )
        self._settle_time = 0

    @property
    def settle_time(self):
        return self._settle_time

    @settle_time.setter
    def settle_time(self, value):
        if value < 0:
            raise ValueError("Settle value needs to be >= 0.")
        else:
            self._settle_time = value

    def pause(self):
        """Change setpoint to current position."""
        self.setpoint.put(self._position)


class LakeShore336_LoopRO(Device):
    """
    LakeShore 336 temperature controller -- Read-only loop (no heaters).
    """

    # Set this to normal because we don't use it.
    readback = FormattedComponent(
        EpicsSignalRO, "{prefix}IN{loop_number}", kind="normal"
    )
    units = FormattedComponent(
        EpicsSignalWithRBV, "{prefix}IN{loop_number}:Units", kind="omitted"
    )
    loop_name = FormattedComponent(
        EpicsSignalRO, "{prefix}IN{loop_number}:Name_RBV", kind="omitted"
    )

    def __init__(self, *args, loop_number=None, **kwargs):
        self.loop_number = loop_number
        super().__init__(*args, **kwargs)


class LakeShore336Device(Device):
    """
    LakeShore 336 temperature controller.

    - loop 1: temperature positioner AND heater, PID, & ramp controls
    - loop 2: temperature positioner AND heater, PID, & ramp controls
    - loop 3: temperature positioner
    - loop 4: temperature positioner
    """

    loop1 = FormattedComponent(
        LakeShore336_LoopControl, "{self.prefix}", loop_number=1
    )
    loop2 = FormattedComponent(
        LakeShore336_LoopControl, "{self.prefix}", loop_number=2
    )
    loop3 = FormattedComponent(
        LakeShore336_LoopRO, "{self.prefix}", loop_number=3
    )
    loop4 = FormattedComponent(
        LakeShore336_LoopRO, "{self.prefix}", loop_number=4
    )

    # same names as apstools.synApps._common.EpicsRecordDeviceCommonAll
    scanning_rate = Component(EpicsSignal, "read.SCAN", kind="omitted")
    process_record = Component(EpicsSignal, "read.PROC", kind="omitted")

    read_all = Component(EpicsSignal, "readAll.PROC", kind="omitted")
    serial = Component(AsynRecord, "serial", kind="omitted")


class LS340_LoopBase(PVPositionerSoftDoneWithStop):
    """ Base settings for both sample and control loops. """

    # position
    # readback is provided by PVPositionerSoftDoneWithStop
    # TODO: Polar reports: sometimes the setpoint PV is not updated
    setpoint = FormattedComponent(
        EpicsSignal,
        "{prefix}SP{loop_number}",
        write_pv="{prefix}wr_SP{loop_number}",
        kind="normal",
        put_complete=True,
    )

    # configuration
    units = Component(Signal, value="K", kind="config")

    pid_P = FormattedComponent(
        EpicsSignal,
        "{prefix}P{loop_number}",
        write_pv="{prefix}setPID{loop_number}.AA",
        kind="config",
    )
    pid_I = FormattedComponent(
        EpicsSignal,
        "{prefix}I{loop_number}",
        write_pv="{prefix}setPID{loop_number}.BB",
        kind="config",
    )
    pid_D = FormattedComponent(
        EpicsSignal,
        "{prefix}D{loop_number}",
        write_pv="{prefix}setPID{loop_number}.CC",
        kind="config",
    )

    ramp_rate = FormattedComponent(
        EpicsSignal,
        "{prefix}Ramp{loop_number}",
        write_pv="{prefix}setRamp{loop_number}.BB",
        kind="config",
    )
    ramp_on = FormattedComponent(
        EpicsSignal, "{prefix}Ramp{loop_number}_on", kind="config"
    )

    def __init__(self, *args, loop_number=None, timeout=10 * HOUR, **kwargs):
        self.loop_number = loop_number
        super().__init__(
            *args,
            readback_pv="ignore",
            timeout=timeout,
            tolerance=0.1,
            **kwargs
        )
        self._settle_time = 0

    @property
    def settle_time(self):
        return self._settle_time

    @settle_time.setter
    def settle_time(self, value):
        if value < 0:
            raise ValueError("Settle value needs to be >= 0.")
        else:
            self._settle_time = value

    @property
    def egu(self):
        return self.units.get(as_string=True)

    def pause(self):
        """Change setpoint to current position."""
        self.setpoint.put(self._position, wait=True)


class LS340_LoopControl(LS340_LoopBase):
    """ Control specific """

    readback = Component(EpicsSignalRO, "Control", kind="normal")
    sensor = Component(EpicsSignal, "Ctl_sel", kind="config")


class LS340_LoopSample(LS340_LoopBase):
    """ Sample specific """

    readback = Component(EpicsSignalRO, "Sample", kind="hinted")
    sensor = Component(EpicsSignal, "Spl_sel", kind="config")


class LakeShore340Device(Device):
    """
    LakeShore 340 temperature controller
    """

    control = FormattedComponent(LS340_LoopControl, "{prefix}", loop_number=1)
    sample = FormattedComponent(LS340_LoopSample, "{prefix}", loop_number=2)

    heater = Component(EpicsSignalRO, "Heater")
    heater_range = Component(
        EpicsSignal,
        "Rg_rdbk",
        write_pv="HeatRg",
        kind="normal",
        put_complete=True
    )

    read_pid = Component(EpicsSignal, "readPID.PROC", kind="omitted")

    # same names as apstools.synApps._common.EpicsRecordDeviceCommonAll
    scanning_rate = Component(EpicsSignal, "read.SCAN", kind="omitted")
    process_record = Component(EpicsSignal, "read.PROC", kind="omitted")

    serial = Component(AsynRecord, "serial", kind="omitted")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



# content of conftest.py

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--run-local", action="store_true", default=False, help="run local tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "local: mark test to run locally")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run-local"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_local = pytest.mark.skip(reason="need --run-local option to run")
    for item in items:
        if "local" in item.keywords:
            item.add_marker(skip_local)



"""
Searching databroker catalogs
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~db_query
"""

import databroker
from ._core import FIRST_DATA
from ._core import LAST_DATA


def db_query(db, query):
    """
    Searches the databroker v2 database.

    PARAMETERS

    db
        *object* :
        Bluesky database, an instance of ``databroker.catalog``.

    query
        *dict* :
        Search parameters.

    RETURNS

    *object* :
        Bluesky database, an instance of ``databroker.catalog``
        satisfying the ``query`` parameters.

    .. seealso::

       :func:`databroker.catalog.search`
    """

    if query is None:
        return db

    since = query.pop("since", None)
    until = query.pop("until", None)

    if since or until:
        if not since:
            since = FIRST_DATA
        if not until:
            until = LAST_DATA

        # fmt: off
        _db = db.v2.search(
            databroker.queries.TimeRange(since=since, until=until)
        )
        # fmt: on
    else:
        _db = db

    if len(query) != 0:
        _db = _db.v2.search(query)

    return _db

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
APS undulator
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~ApsUndulator
   ~ApsUndulatorDual
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import Signal
from .tracking_signal import TrackingSignal


class ApsUndulator(Device):
    """
    APS Undulator

    .. index:: Ophyd Device; ApsUndulator

    EXAMPLE::

        undulator = ApsUndulator("ID09ds:", name="undulator")
    """

    energy = Component(
        EpicsSignal, "Energy", write_pv="EnergySet", put_complete=True, kind="hinted",
    )
    energy_taper = Component(
        EpicsSignal, "TaperEnergy", write_pv="TaperEnergySet", kind="config",
    )
    gap = Component(EpicsSignal, "Gap", write_pv="GapSet")
    gap_taper = Component(
        EpicsSignal, "TaperGap", write_pv="TaperGapSet", kind="config"
    )
    start_button = Component(EpicsSignal, "Start", put_complete=True, kind="omitted")
    stop_button = Component(EpicsSignal, "Stop", kind="omitted")
    harmonic_value = Component(EpicsSignal, "HarmonicValue", kind="config")
    gap_deadband = Component(EpicsSignal, "DeadbandGap", kind="config")
    device_limit = Component(EpicsSignal, "DeviceLimit", kind="config")

    access_mode = Component(EpicsSignalRO, "AccessSecurity", kind="omitted")
    device_status = Component(EpicsSignalRO, "Busy", kind="omitted")
    total_power = Component(EpicsSignalRO, "TotalPower", kind="config")
    message1 = Component(EpicsSignalRO, "Message1", kind="omitted")
    message2 = Component(EpicsSignalRO, "Message2", kind="omitted")
    message3 = Component(EpicsSignalRO, "Message3", kind="omitted")
    time_left = Component(EpicsSignalRO, "ShClosedTime", kind="omitted")

    device = Component(EpicsSignalRO, "Device", kind="config")
    location = Component(EpicsSignalRO, "Location", kind="config")
    version = Component(EpicsSignalRO, "Version", kind="config")

    # Useful undulator parameters that are not EPICS PVs.
    energy_deadband = Component(Signal, value=0.0, kind="config")
    energy_backlash = Component(Signal, value=0.0, kind="config")
    energy_offset = Component(Signal, value=0, kind="config")
    tracking = Component(TrackingSignal, value=False, kind="config")


class ApsUndulatorDual(Device):
    """
    APS Undulator with upstream *and* downstream controls

    .. index:: Ophyd Device; ApsUndulatorDual

    EXAMPLE::

        undulator = ApsUndulatorDual("ID09", name="undulator")

    note:: the trailing ``:`` in the PV prefix should be omitted
    """

    upstream = Component(ApsUndulator, "us:")
    downstream = Component(ApsUndulator, "ds:")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Mixin to add EPICS .DESC field
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~EpicsDescriptionMixin
"""

from ophyd import Component
from ophyd import EpicsSignal
from .mixin_base import DeviceMixinBase


class EpicsDescriptionMixin(DeviceMixinBase):
    """
    add a record's description field to a Device, such as EpicsMotor

    .. index:: Ophyd Device Mixin; EpicsDescriptionMixin

    EXAMPLE::

        from ophyd import EpicsMotor
        from apstools.devices import EpicsDescriptionMixin

        class MyEpicsMotor(EpicsDescriptionMixin, EpicsMotor): pass
        m1 = MyEpicsMotor('xxx:m1', name='m1')
        print(m1.desc.get())

    more ideas::

        class TunableSynAxis(AxisTunerMixin, SynAxis):
            '''synthetic axis that can be tuned'''
        class TunableEpicsMotor(AxisTunerMixin, EpicsMotor):
            '''EpicsMotor that can be tuned'''
        class EpicsMotorWithDescription(EpicsDescriptionMixin, EpicsMotor):
            '''EpicsMotor with description field'''

        class EpicsMotorWithMore(
            EpicsDescriptionMixin,
            EpicsMotorLimitsMixin,
            EpicsMotorDialMixin,
            EpicsMotorRawMixin,
            EpicsMotor):
            '''
            EpicsMotor with more fields

            * description (``desc``)
            * soft motor limits (``soft_limit_hi``, ``soft_limit_lo``)
            * dial coordinates (``dial``)
            * raw coordinates (``raw``)
            '''
    """

    desc = Component(EpicsSignal, ".DESC")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Kohzu double-crystal monochromator
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~KohzuSeqCtl_Monochromator
"""

import logging
import time

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import Signal
from ophyd import PVPositioner

from ..utils import run_in_thread

logger = logging.getLogger(__name__)


class KohzuSoftPositioner(PVPositioner):
    setpoint = Component(EpicsSignal, "AO", kind="normal")
    readback = Component(EpicsSignalRO, "RdbkAO", kind="hinted")
    done = Component(Signal, value=0, kind="omitted")
    done_value = 0

    def cb_done(self, *args, **kwargs):
        """
        Called when parent's done signal changes (EPICS CA monitor event).
        """
        self.done.put(self.parent.moving.get())

    def cb_setpoint(self, *args, **kwargs):
        """
        Called when setpoint changes (EPICS CA monitor event).

        When the setpoint is changed, force ``done=False``.  For any move,
        done must transition to ``!= done_value``, then back to ``done_value``.
        Next update will refresh value from parent device.
        """
        self.done.put(1)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Prepare the positioner to be stopped.
        # Tells monochromator to stop its motors.  Brutal.
        self.stop_signal = self.parent.allstop_button
        self.stop_value = 1

        # setup callbacks on done and setpoint
        self.setpoint.subscribe(self.cb_setpoint)
        self.parent.moving.subscribe(self.cb_done)

        # the readback needs no adjective
        self.readback.name = self.name

    @property
    def inposition(self):
        """Report (boolean) if positioner is done."""
        return self.done.get() == self.done_value

    def move(self, *args, **kwargs):
        """Reposition, with optional wait for completion."""
        @run_in_thread
        def push_the_move_button_soon(delay_s=0.01):
            time.sleep(delay_s)  # wait a short time
            self.parent.move_button.put(1)

        push_the_move_button_soon()
        return super().move(*args, **kwargs)

    def _setup_move(self, position):
        """Move and do not wait until motion is complete (asynchronous)."""
        self.log.debug("%s.setpoint = %s", self.name, position)
        self.setpoint.put(position, wait=False)  # was: wait=True
        if self.actuate is not None:
            self.log.debug("%s.actuate = %s", self.name, self.actuate_value)
            self.actuate.put(self.actuate_value, wait=False)


class KohzuSeqCtl_Monochromator(Device):
    """
    synApps Kohzu double-crystal monochromator sequence control program

    .. index:: Ophyd Device; KohzuSeqCtl_Monochromator
    """

    energy = Component(KohzuSoftPositioner, "BraggE", kind="hinted")
    theta = Component(KohzuSoftPositioner, "BraggTheta", kind="normal")
    # lambda is reserved word in Python, can't use it for wavelength
    wavelength = Component(KohzuSoftPositioner, "BraggLambda", kind="normal")

    y1 = Component(EpicsSignalRO, "KohzuYRdbkAI", kind="normal")
    z2 = Component(EpicsSignalRO, "KohzuZRdbkAI", kind="normal")

    message2 = Component(EpicsSignalRO, "KohzuSeqMsg2SI", kind="config")
    operator_acknowledge = Component(EpicsSignal, "KohzuOperAckBO", kind="omitted")
    use_set = Component(EpicsSignal, "KohzuUseSetBO", kind="omitted")
    mode = Component(EpicsSignal, "KohzuModeBO", kind="config", string=True)
    move_button = Component(
        EpicsSignal, "KohzuPutBO", put_complete=True, kind="omitted"
    )
    moving = Component(EpicsSignal, "KohzuMoving", kind="omitted")
    y_offset = Component(EpicsSignal, "Kohzu_yOffsetAO", kind="config")
    allstop_button = Component(EpicsSignal, "allstop", string=True, kind="omitted")

    crystal_mode = Component(EpicsSignal, "KohzuMode2MO", string=True, kind="config")
    crystal_h = Component(EpicsSignal, "BraggHAO", kind="config")
    crystal_k = Component(EpicsSignal, "BraggKAO", kind="config")
    crystal_l = Component(EpicsSignal, "BraggLAO", kind="config")
    crystal_lattice_constant = Component(EpicsSignal, "BraggAAO", kind="config")
    crystal_2d_spacing = Component(EpicsSignal, "Bragg2dSpacingAO", kind="config")
    crystal_type = Component(EpicsSignal, "BraggTypeMO", string=True, kind="config")

    def calibrate_energy(self, value):
        """Calibrate the monochromator energy.

        PARAMETERS

        value: float
            New energy for the current monochromator position.
        """
        self.use_set.put("Set")
        self.energy.put(value)
        self.use_set.put("Use")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
sscan Record plans
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~sscan_1D
"""

from collections import OrderedDict
import time

from bluesky import plan_stubs as bps
from ophyd import DeviceStatus


new_data = False  # sscan has new data, boolean
inactive_deadline = 0  # sscan timeout, absolute time.time()


def _get_sscan_data_objects(sscan):
    """
    prepare a dictionary of the "interesting" ophyd data objects for this sscan

    PARAMETERS

    sscan
        *Device* :
        one EPICS sscan record (instance of `apstools.synApps.sscanRecord`)

    """
    scan_data_objects = OrderedDict()
    for part in (sscan.positioners, sscan.detectors):
        for chname in part.read_attrs:
            if not chname.endswith("_value"):
                continue
            obj = getattr(part, chname)
            key = obj.name.lstrip(sscan.name + "_")
            scan_data_objects[key] = obj
    return scan_data_objects


def sscan_1D(
    sscan,
    poll_delay_s=0.001,
    phase_timeout_s=60.0,
    running_stream="primary",
    final_array_stream=None,
    device_settings_stream="settings",
    md=None,
):
    """
    simple 1-D scan using EPICS synApps sscan record

    .. index:: Bluesky Plan; sscan_1D

    assumes the sscan record has already been setup properly for a scan

    PARAMETERS

    sscan *Device* :
        one EPICS sscan record (instance of `apstools.synApps.sscanRecord`)
    running_stream *str* : or `None`
        (default: ``"primary"``)
        Name of document stream to write positioners and detectors data
        made available while the sscan is running.  This is typically
        the scan data, row by row.
        If set to `None`, this stream will not be written.
    final_array_stream *str*  or ``None`` :
        Name of document stream to write positioners and detectors data
        posted *after* the sscan has ended.
        If set to `None`, this stream will not be written.
        (default: ``None``)
    device_settings_stream *str*  or ``None`` :
        Name of document stream to write *settings* of the sscan device.
        This is all the information returned by ``sscan.read()``.
        If set to `None`, this stream will not be written.
        (default: ``"settings"``)
    poll_delay_s *float* :
        How long to sleep during each polling loop while collecting
        interim data values and waiting for sscan to complete.
        Must be a number between zero and 0.1 seconds.
        (default: 0.001 seconds)
    phase_timeout_s *float* :
        How long to wait after last update of the ``sscan.FAZE``.
        When scanning, we expect the scan phase to update regularly
        as positioners move and detectors are triggered.  If the scan
        hangs for some reason, this is a way to end the plan early.
        To cancel this feature, set it to ``None``.
        (default: 60 seconds)

    NOTE about the document stream names

    Make certain the names for the document streams are different from
    each other.  If you make them all the same (such as ``primary``),
    you will have difficulty when reading your data later on.

    *Don't cross the streams!*

    EXAMPLE

    Assume that the chosen sscan record has already been setup.

        from apstools.devices import sscanDevice
        scans = sscanDevice(P, name="scans")

        from apstools.plans import sscan_1D
        RE(sscan_1D(scans.scan1), md=dict(purpose="demo"))

    """
    global new_data, inactive_deadline

    if not (0 <= poll_delay_s <= 0.1):
        raise ValueError(
            "poll_delay_s must be a number between 0 and 0.1,"
            f" received {poll_delay_s}"
        )

    t0 = time.time()
    sscan_status = DeviceStatus(sscan.execute_scan)
    started = False
    new_data = False
    inactive_deadline = time.time()
    if phase_timeout_s is not None:
        inactive_deadline += phase_timeout_s

    def execute_cb(value, timestamp, **kwargs):
        """watch for sscan to complete"""
        if started and value in (0, "IDLE"):
            sscan_status._finished()
            sscan.execute_scan.unsubscribe_all()
            sscan.scan_phase.unsubscribe_all()

    def phase_cb(value, timestamp, **kwargs):
        """watch for new data"""
        global new_data, inactive_deadline
        if phase_timeout_s is not None:
            inactive_deadline = time.time() + phase_timeout_s
        if value in (15, "RECORD SCALAR DATA"):
            new_data = True  # set flag for main plan

    # acquire only the channels with non-empty configuration in EPICS
    sscan.select_channels()
    # pre-identify the configured channels
    sscan_data_objects = _get_sscan_data_objects(sscan)

    # watch for sscan to complete
    sscan.execute_scan.subscribe(execute_cb)
    # watch for new data to be read out
    sscan.scan_phase.subscribe(phase_cb)

    _md = dict(plan_name="sscan_1D")
    _md.update(md or {})

    uid = yield from bps.open_run(_md)  # start data collection
    yield from bps.mv(sscan.execute_scan, 1)  # start sscan
    started = True

    # collect and emit data, wait for sscan to end
    while not sscan_status.done or new_data:
        if new_data and running_stream is not None:
            yield from bps.create(running_stream)
            for _k, obj in sscan_data_objects.items():
                yield from bps.read(obj)
            yield from bps.save()
        new_data = False
        if phase_timeout_s is not None and time.time() > inactive_deadline:
            print(
                f"No change in sscan record for {phase_timeout_s} seconds."
            )
            print("ending plan early as unsuccessful")
            sscan_status._finished(success=False)
        yield from bps.sleep(poll_delay_s)

    # dump the complete data arrays
    if final_array_stream is not None:
        yield from bps.create(final_array_stream)
        # we have to search for the arrays since they have ``kind="omitted"``
        # (which means they do not get reported by the ``.read()`` method)
        for part in (sscan.positioners, sscan.detectors):
            for nm in part.read_attrs:
                if "." not in nm:
                    # TODO: write just the acquired data, not the FULL arrays!
                    yield from bps.read(getattr(part, nm).array)
        yield from bps.save()

    # dump the entire sscan record into another stream
    if device_settings_stream is not None:
        yield from bps.create(device_settings_stream)
        yield from bps.read(sscan)
        yield from bps.save()

    yield from bps.close_run()

    elapsed = time.time() - t0
    print(f"total time for sscan_1D: {elapsed} s")

    return uid

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



import time
import warnings

IOC = "gp:"
MASTER_TIMEOUT = 30
MAX_TESTING_RETRIES = 3
SHORT_DELAY_FOR_EPICS_IOC_DATABASE_PROCESSING = 2. / 60  # two 60Hz clock cycles


def short_delay_for_EPICS_IOC_database_processing(delay=None):
    time.sleep(delay or SHORT_DELAY_FOR_EPICS_IOC_DATABASE_PROCESSING)


def common_attribute_quantities_test(device, pv, connect, attr, expected):
    """
    Verify the quantities of an attribute.

    Count the number (or number of lines) of the named attribute. Special cases
    for certain methods.  Here are the known cases:

        * configuration_attrs
        * read_attrs
        * read()
        * summary()

    PARAMETERS

    device class:
        The class to be constructed for testing.
    pv str:
        The EPICS Process Variable name (or prefix) to be used.
    connect bool:
        Does testing require the object's Components to be connected with EPICS?
    attr str:
        The attribute (or method) to be tested.  Unhandled methods will raise an
        AttributeError.
    expected int:
        Expected quantity:  ``len(attr)`` (see code below)

    NOTE:  A retry handler was implemented to respond to some cases where random
    Components failed to connect with the default timeout. Testing with longer
    timeout intervals did not succeed while retrying the initial connection was
    successful.  A warning is reported when a timeout is encountered.
    """
    retry = 0
    while retry < MAX_TESTING_RETRIES:
        retry += 1
        obj = device(pv, name="obj")
        assert obj is not None, pv
        if not connect:
            break  # no need to wait
        try:
            obj.wait_for_connection()
            break
        except TimeoutError:
            # note: Increasing interval did not resolve the timeout.
            # Forcing retries was successful
            # This problem is intermittent.
            warnings.warn(
                f"Timeout connecting {attr} in {retry}/{MAX_TESTING_RETRIES}"
            )

    if connect:
        assert obj.connected, f"{pv} {attr} {connect}"

    if attr == "read()":
        l = len(obj.read())
    elif attr == "summary()":
        l = len(obj._summary().splitlines())
    else:
        l = len(getattr(obj, attr))
    assert l == expected, f"{attr}: {l} != {expected}"



"""
Area Detector Support
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~AD_EpicsFileNameHDF5Plugin
   ~AD_EpicsFileNameJPEGPlugin
   ~AD_EpicsFileNameMixin
   ~AD_EpicsFileNameTIFFPlugin
   ~AD_EpicsHdf5FileName
   ~AD_EpicsHDF5IterativeWriter
   ~AD_EpicsJPEGFileName
   ~AD_EpicsJPEGIterativeWriter
   ~AD_EpicsTIFFFileName
   ~AD_EpicsTIFFIterativeWriter
   ~AD_full_file_name_local
   ~AD_plugin_primed
   ~AD_prime_plugin
   ~AD_prime_plugin2
   ~AD_setup_FrameType
   ~CamMixin_V3_1_1
   ~CamMixin_V34
   ~SingleTrigger_V34
"""

from collections import OrderedDict
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import EpicsSignalWithRBV
from ophyd.areadetector import ADComponent
from ophyd.areadetector import CamBase
from ophyd.areadetector import SingleTrigger
from ophyd.areadetector.filestore_mixins import FileStoreBase
from ophyd.areadetector.filestore_mixins import FileStoreIterativeWrite
from ophyd.areadetector.filestore_mixins import FileStorePluginBase
from ophyd.areadetector.plugins import HDF5Plugin_V34 as HDF5Plugin
from ophyd.areadetector.plugins import JPEGPlugin_V34 as JPEGPlugin
from ophyd.areadetector.plugins import TIFFPlugin_V34 as TIFFPlugin
import datetime
import epics
import itertools
import logging
import numpy as np
from packaging import version
import pathlib
import time
import warnings

logger = logging.getLogger(__name__)

# fmt: off
AD_FrameType_schemes = {
    "reset": dict(  # default names from Area Detector code
        ZRST="Normal",
        ONST="Background",
        TWST="FlatField",
    ),
    "NeXus": dict(  # NeXus (typical locations)
        ZRST="/entry/data/data",
        ONST="/entry/data/dark",
        TWST="/entry/data/white",
    ),
    "DataExchange": dict(  # APS Data Exchange
        ZRST="/exchange/data",
        ONST="/exchange/data_dark",
        TWST="/exchange/data_white",
    ),
}
# fmt: on


def AD_setup_FrameType(prefix, scheme="NeXus"):
    """
    configure so frames are identified & handled by type (dark, white, or image)

    PARAMETERS

        prefix
            *str* :
            EPICS PV prefix of area detector, such as ``13SIM1:``
        scheme
            *str* :
            any key in the ``AD_FrameType_schemes`` dictionary

    This routine prepares the EPICS Area Detector to identify frames
    by image type for handling by clients, such as the HDF5 file writing plugin.
    With the HDF5 plugin, the ``FrameType`` PV is added to the NDattributes
    and then used in the layout file to direct the acquired frame to
    the chosen dataset.  The ``FrameType`` PV value provides the HDF5 address
    to be used.

    To use a different scheme than the defaults, add a new key to
    the ``AD_FrameType_schemes`` dictionary, defining storage values for the
    fields of the EPICS ``mbbo`` record that you will be using.

    see: https://nbviewer.jupyter.org/github/BCDA-APS/use_bluesky/blob/main/lessons/sandbox/images_darks_flats.ipynb

    EXAMPLE::

        AD_setup_FrameType("2bmbPG3:", scheme="DataExchange")

    * Call this function *before* creating the ophyd area detector object
    * use lower-level PyEpics interface
    """
    db = AD_FrameType_schemes.get(scheme)
    if db is None:
        raise ValueError(
            f"unknown AD_FrameType_schemes scheme: {scheme}"
            "\n Should be one of: " + ", ".join(AD_FrameType_schemes.keys())
        )

    template = "{}cam1:FrameType{}.{}"
    for field, value in db.items():
        epics.caput(template.format(prefix, "", field), value)
        epics.caput(template.format(prefix, "_RBV", field), value)


def AD_plugin_primed(plugin):
    """
    Has area detector pushed an NDarray to the file writer plugin?  True or False

    PARAMETERS

    plugin
        *obj* :
        area detector plugin to be *primed* (such as ``detector.hdf1``)

    EXAMPLE::

        AD_plugin_primed(detector.hdf1)

    Works around an observed issue: #598
    https://github.com/NSLS-II/ophyd/issues/598#issuecomment-414311372

    If detector IOC has just been started and has not yet taken an image
    with the file writer plugin, then a TimeoutError will occur as the
    file writer plugin "Capture" is set to 1 (Start).  In such case,
    first acquire at least one image with the file writer plugin enabled.

    Also issue in apstools (needs a robust method to detect if primed):
    https://github.com/BCDA-APS/apstools/issues/464

    Since Area Detector release 2.1 (2014-10-14).

    The *prime* process is not needed if you select the
    *LazyOpen* feature with *Stream* mode for the file plugin.
    *LazyOpen* defers file creation until the first frame arrives
    in the plugin. This removes the need to initialize the plugin
    with a dummy frame before starting capture.
    """
    cam = plugin.parent.cam
    tests = []

    for obj in (cam, plugin):
        test = np.array(obj.array_size.get()).sum() != 0
        tests.append(test)
        if not test:
            logger.debug("'%s' image size is zero", obj.name)

    checks = dict(
        array_size=False,
        color_mode=True,
        data_type=True,
    )
    for key, as_string in checks.items():
        c = getattr(cam, key).get(as_string=as_string)
        p = getattr(plugin, key).get(as_string=as_string)
        test = c == p
        tests.append(test)
        if not test:
            logger.debug("%s does not match", key)

    return False not in tests


def AD_prime_plugin(detector, plugin):
    """
    Prime this area detector's file writer plugin.

    PARAMETERS

    detector
        *obj* :
        area detector (such as ``detector``)
    plugin
        *obj* :
        area detector plugin to be *primed* (such as ``detector.hdf1``)

    EXAMPLE::

        AD_prime_plugin(detector, detector.hdf1)
    """
    nm = f"{plugin.parent.name}.{plugin.attr_name}"
    warnings.warn(f"Use AD_prime_plugin2({nm}) instead.")
    AD_prime_plugin2(plugin)


def AD_prime_plugin2(plugin):
    """
    Prime this area detector's file writer plugin.

    Collect and push an NDarray to the file writer plugin.
    Works with all file writer plugins.

    Based on ``ophyd.areadetector.plugins.HDF5Plugin.warmup()``.

    PARAMETERS

    plugin
        *obj* :
        area detector plugin to be *primed* (such as ``detector.hdf1``)

    EXAMPLE::

        AD_prime_plugin2(detector.hdf1)

    """
    if AD_plugin_primed(plugin):
        logger.debug("'%s' plugin is already primed", plugin.name)
        return

    sigs = OrderedDict(
        [
            (plugin.enable, 1),
            (plugin.parent.cam.array_callbacks, 1),  # set by number
            (plugin.parent.cam.image_mode, 0),  # Single, set by number
            # Trigger mode names are not identical for every camera.
            # Assume here that the first item in the list is
            # the best default choice to prime the plugin.
            (plugin.parent.cam.trigger_mode, 0),  # set by number
            # just in case the acquisition time is set very long...
            (plugin.parent.cam.acquire_time, 1),
            (plugin.parent.cam.acquire_period, 1),
            (plugin.parent.cam.acquire, 1),  # set by number
        ]
    )

    original_vals = {sig: sig.get() for sig in sigs}

    for sig, val in sigs.items():
        time.sleep(0.1)  # abundance of caution
        sig.set(val).wait()

    time.sleep(2)  # wait for acquisition

    for sig, val in reversed(list(original_vals.items())):
        time.sleep(0.1)
        sig.set(val).wait()


def AD_full_file_name_local(plugin):
    """
    Return AD plugin's *Last filename* using local filesystem path.

    Get the full name, in terms of the bluesky filesystem, for the image file
    recently-acquired by the area detector plugin.

    Return the name as a pathlib object.

    PARAMETERS

    plugin *obj* :
        Instance of ophyd area detector file writing plugin.

    (new in apstools release 1.6.2)
    """
    fname = plugin.full_file_name.get().strip()
    if fname == "":
        return None

    ffname = pathlib.Path(fname)  # FIXME: OS style?
    if plugin.read_path_template == plugin.write_path_template:
        return ffname

    # identify the common last parts of the file directories
    read_parts = pathlib.Path(plugin.read_path_template).parts
    write_parts = pathlib.Path(plugin.write_path_template).parts
    icommon = 0
    for i in range(min(len(read_parts), len(write_parts))):
        i1 = -i - 1
        if read_parts[i1:] != write_parts[i1:]:
            icommon = i
            break

    if icommon == 0:
        raise ValueError(
            "No common part to file paths. " "Cannot convert to local file path."
        )

    # fmt: off
    local_root = pathlib.Path().joinpath(*read_parts[:-icommon])
    common_parts = ffname.parts[len(write_parts[:-icommon]):]
    local_ffname = local_root.joinpath(*common_parts)
    # fmt: on

    return local_ffname


class AD_EpicsFileNameMixin(FileStorePluginBase):
    """
    Custom class to define image file name from EPICS.

    Used as part of AD_EpicsFileNameHDF5Plugin.

    .. index:: Ophyd Device Support; AD_EpicsHdf5FileName

    .. caution:: *Caveat emptor* applies here.  You assume expertise!

    Replace standard ophyd file naming algorithm (where file names
    are defined as UUID strings, virtually guaranteeing that
    no existing images files will ever be overwritten).

    Caller is responsible for setting values of these Components:

    * array_counter
    * auto_increment
    * auto_save
    * compression (only HDF)
    * create_directory
    * file_name
    * file_number
    * file_path
    * file_template
    * num_capture

    .. autosummary::

        ~make_filename
        ~get_frames_per_point
        ~stage

    To allow users to control the file **name**,
    we override the ``make_filename()`` method here
    and we need to override some intervening classes.

    To allow users to control the file **number**,
    we override the ``stage()`` method here
    and triple-comment out that line, and bring in
    sections from the methods we are replacing here.

    It is allowed to set the ``file_template="%s%s.h5"``
    so the file name does not include the file number.

    The image file name is set in ``FileStoreBase.make_filename()``
    from ``ophyd.areadetector.filestore_mixins``.  This is called
    (during device staging) from ``FileStoreBase.stage()``
    """

    def _remove_caller_stage_sigs(self):
        """Caller is responsible for setting these stage_sigs."""
        caller_sets_these = """
        array_counter
        auto_increment
        auto_save
        compression (only HDF)
        create_directory
        file_name
        file_number
        file_path
        file_template
        num_capture
        """.split()
        for key in caller_sets_these:
            if key in self.stage_sigs:
                self.stage_sigs.pop(key)

    def make_filename(self):
        """
        overrides default behavior: Get info from EPICS file writer plugin.
        """
        # start of the file name, file number will be appended per template
        filename = self.file_name.get()
        file_path = self.file_path.get()
        formatter = datetime.datetime.now().strftime

        # Directory (used by IOC) for file writer plugin to write the file.
        write_path = formatter(file_path)

        # Directory (used by bluesky) for databroker to read the file.
        read_path = formatter(self.read_path_template)

        return filename, read_path, write_path

    def get_frames_per_point(self):
        """overrides default behavior"""
        return self.num_capture.get()

    def stage(self):
        """
        Overrides default behavior of parent class.

        Parent class items overridden here:

        * Sets file_name based on a UUID.
        * Sets file_path from write_path_template.
        * Sets file_number to 0.

        Set EPICS items before device is staged, then copy EPICS
        naming template (and other items) to ophyd after staging.
        """
        if "capture" in self.stage_sigs:
            self.stage_sigs.move_to_end("capture", last=True)

        # Get the file name and paths from EPICS.
        filename, read_path, write_path = self.make_filename()

        # Ensure we do not have an old file open.
        self.capture.set(0).wait()

        # Set these before capture is turned on.
        # They will not be reset on 'unstage' anyway.
        self.file_path.set(write_path).wait()
        self.file_name.set(filename).wait()

        # Get file number now, it is incremented during stage().
        file_number = self.file_number.get()

        # Call ancestor's stage(), skipping parent's stage().
        FileStoreBase.stage(self)

        # AD applies the file name templating in C.
        # We can't access that result until after acquisition
        # so we apply the same template here in Python.
        template = self.file_template.get(use_monitor=False)
        try:
            # assume template includes format for a file number
            self._fn = template % (read_path, filename, file_number)
        except (TypeError, ValueError):
            # in case template does not include file_number
            self._fn = template % (read_path, filename)
        self._fp = read_path
        if not self.file_path_exists.get():
            raise IOError(f"Path '{self.file_path.get()}' does not exist on IOC.")

        # index each image frame (used in generate_datum() method)
        self._point_counter = itertools.count()

        # from FileStoreHDF5.stage()
        res_kwargs = {"frame_per_point": self.get_frames_per_point()}
        self._generate_resource(res_kwargs)


class AD_EpicsHdf5FileName(AD_EpicsFileNameMixin):
    """
    Custom class to define HDF5 image file name from EPICS PVs.

    Used as part of AD_EpicsFileNameHDF5Plugin.
    """

    def __init__(self, *args, **kwargs):
        FileStorePluginBase.__init__(self, *args, **kwargs)
        self.filestore_spec = "AD_HDF5"  # spec name stored in resource doc
        self.stage_sigs.update(
            [
                ("file_write_mode", "Stream"),
                ("capture", 1),
            ]
        )
        self._remove_caller_stage_sigs()
        # "capture" must always come last
        self.stage_sigs["capture"] = self.stage_sigs.pop("capture")


class AD_EpicsHDF5IterativeWriter(AD_EpicsHdf5FileName, FileStoreIterativeWrite):
    """
    intermediate class between AD_EpicsHdf5FileName and AD_EpicsFileNameHDF5Plugin

    (new in apstools release 1.6.2)
    """

    pass


class AD_EpicsFileNameHDF5Plugin(HDF5Plugin, AD_EpicsHDF5IterativeWriter):
    """
    Alternative to HDF5Plugin: EPICS area detector PV sets file name.

    .. index:: Ophyd Device Support; AD_EpicsFileNameHDF5Plugin

    .. caution:: *Caveat emptor* applies here.  You assume expertise!

    Uses ``AD_EpicsHdf5FileName``.

    EXAMPLE::

        from apstools.devices.area_detector_support import AD_EpicsFileNameHDF5Plugin
        from ophyd import EpicsSignalWithRBV
        from ophyd.areadetector import ADComponent
        from ophyd.areadetector import DetectorBase
        from ophyd.areadetector import SingleTrigger
        from ophyd.areadetector.plugins import ImagePlugin_V34 as ImagePlugin
        from ophyd.areadetector.plugins import PvaPlugin_V34 as PvaPlugin
        from ophyd.areadetector import SimDetectorCam
        import datetime
        import pathlib


        IOC = "ad:"
        IMAGE_DIR = "adsimdet/%Y/%m/%d"
        AD_IOC_MOUNT_PATH = pathlib.Path("/tmp")
        BLUESKY_MOUNT_PATH = pathlib.Path("/tmp/docker_ioc/iocad/tmp")

        # MUST end with a `/`, pathlib will NOT provide it
        WRITE_PATH_TEMPLATE = f"{AD_IOC_MOUNT_PATH / IMAGE_DIR}/"
        READ_PATH_TEMPLATE = f"{BLUESKY_MOUNT_PATH / IMAGE_DIR}/"


        class MyFixedCam(SimDetectorCam):
            pool_max_buffers = None
            offset = ADComponent(EpicsSignalWithRBV, "Offset")


        class MySimDetector(SingleTrigger, DetectorBase):
            '''ADSimDetector'''

            cam = ADComponent(MyFixedCam, "cam1:")
            image = ADComponent(ImagePlugin, "image1:")
            hdf1 = ADComponent(
                AD_EpicsFileNameHDF5Plugin,
                "HDF1:",
                write_path_template=WRITE_PATH_TEMPLATE,
                read_path_template=READ_PATH_TEMPLATE,
            )
            pva = ADComponent(PvaPlugin, "Pva1:")

    (new in apstools release 1.6.2)
    """

    pass


class AD_EpicsJPEGFileName(AD_EpicsFileNameMixin):
    """
    Custom class to define JPEG image file name from EPICS PVs.

    Used as part of AD_EpicsFileNameJPEGPlugin.
    """

    def __init__(self, *args, **kwargs):
        FileStorePluginBase.__init__(self, *args, **kwargs)
        self.filestore_spec = "AD_JPEG"  # spec name stored in resource doc
        self.stage_sigs.update(
            [
                ("file_write_mode", "Stream"),
                ("capture", 1),
            ]
        )
        self._remove_caller_stage_sigs()
        # "capture" must always come last
        self.stage_sigs["capture"] = self.stage_sigs.pop("capture")


class AD_EpicsJPEGIterativeWriter(AD_EpicsJPEGFileName, FileStoreIterativeWrite):
    """
    intermediate class between AD_EpicsJPEGFileName and AD_EpicsFileNameJPEGPlugin

    (new in apstools release 1.6.2)
    """

    pass


class AD_EpicsFileNameJPEGPlugin(JPEGPlugin, AD_EpicsJPEGIterativeWriter):
    """
    Alternative to JPEGPlugin: EPICS area detector PV sets file name.

    .. index:: Ophyd Device Support; AD_EpicsFileNameJPEGPlugin

    .. caution:: *Caveat emptor* applies here.  You assume expertise!

    Uses ``AD_EpicsJpegFileName``.

    EXAMPLE::

        from apstools.devices.area_detector_support import AD_EpicsFileNameJPEGPlugin
        from ophyd import EpicsSignalWithRBV
        from ophyd.areadetector import ADComponent
        from ophyd.areadetector import DetectorBase
        from ophyd.areadetector import SingleTrigger
        from ophyd.areadetector.plugins import ImagePlugin_V34 as ImagePlugin
        from ophyd.areadetector.plugins import PvaPlugin_V34 as PvaPlugin
        from ophyd.areadetector import SimDetectorCam
        import datetime
        import pathlib


        IOC = "ad:"
        IMAGE_DIR = "adsimdet/%Y/%m/%d"
        AD_IOC_MOUNT_PATH = pathlib.Path("/tmp")
        BLUESKY_MOUNT_PATH = pathlib.Path("/tmp/docker_ioc/iocad/tmp")

        # MUST end with a `/`, pathlib will NOT provide it
        WRITE_PATH_TEMPLATE = f"{AD_IOC_MOUNT_PATH / IMAGE_DIR}/"
        READ_PATH_TEMPLATE = f"{BLUESKY_MOUNT_PATH / IMAGE_DIR}/"


        class MyFixedCam(SimDetectorCam):
            pool_max_buffers = None
            offset = ADComponent(EpicsSignalWithRBV, "Offset")


        class MySimDetector(SingleTrigger, DetectorBase):
            '''ADSimDetector'''

            cam = ADComponent(MyFixedCam, "cam1:")
            image = ADComponent(ImagePlugin, "image1:")
            jpeg1 = ADComponent(
                AD_EpicsFileNameHDF5Plugin,
                "JPEG1:",
                write_path_template=WRITE_PATH_TEMPLATE,
                read_path_template=READ_PATH_TEMPLATE,
            )
            pva = ADComponent(PvaPlugin, "Pva1:")

    (new in apstools release 1.6.2)
    """

    pass


class AD_EpicsTIFFFileName(AD_EpicsFileNameMixin):
    """
    Custom class to define TIFF image file name from EPICS PVs.

    Used as part of AD_EpicsFileNameTIFFPlugin.
    """

    def __init__(self, *args, **kwargs):
        FileStorePluginBase.__init__(self, *args, **kwargs)
        self.filestore_spec = "AD_TIFF"  # spec name stored in resource doc
        self.stage_sigs.update(
            [
                ("file_write_mode", "Stream"),
                ("capture", 1),
            ]
        )
        self._remove_caller_stage_sigs()
        # "capture" must always come last
        self.stage_sigs["capture"] = self.stage_sigs.pop("capture")


class AD_EpicsTIFFIterativeWriter(AD_EpicsTIFFFileName, FileStoreIterativeWrite):
    """
    intermediate class between AD_EpicsTIFFFileName and AD_EpicsFileNameTIFFPlugin

    (new in apstools release 1.6.2)
    """

    pass


class AD_EpicsFileNameTIFFPlugin(TIFFPlugin, AD_EpicsTIFFIterativeWriter):
    """
    Alternative to TIFFPlugin: EPICS area detector PV sets file name.

    .. index:: Ophyd Device Support; AD_EpicsFileNameTIFFPlugin

    .. caution:: *Caveat emptor* applies here.  You assume expertise!

    Uses ``AD_EpicsTIFFFileName``.

    EXAMPLE::

        from apstools.devices.area_detector_support import AD_EpicsFileNameTIFFPlugin
        from ophyd import EpicsSignalWithRBV
        from ophyd.areadetector import ADComponent
        from ophyd.areadetector import DetectorBase
        from ophyd.areadetector import SingleTrigger
        from ophyd.areadetector.plugins import ImagePlugin_V34 as ImagePlugin
        from ophyd.areadetector.plugins import PvaPlugin_V34 as PvaPlugin
        from ophyd.areadetector import SimDetectorCam
        import datetime
        import pathlib


        IOC = "ad:"
        IMAGE_DIR = "adsimdet/%Y/%m/%d"
        AD_IOC_MOUNT_PATH = pathlib.Path("/tmp")
        BLUESKY_MOUNT_PATH = pathlib.Path("/tmp/docker_ioc/iocad/tmp")

        # MUST end with a `/`, pathlib will NOT provide it
        WRITE_PATH_TEMPLATE = f"{AD_IOC_MOUNT_PATH / IMAGE_DIR}/"
        READ_PATH_TEMPLATE = f"{BLUESKY_MOUNT_PATH / IMAGE_DIR}/"


        class MyFixedCam(SimDetectorCam):
            pool_max_buffers = None
            offset = ADComponent(EpicsSignalWithRBV, "Offset")


        class MySimDetector(SingleTrigger, DetectorBase):
            '''ADSimDetector'''

            cam = ADComponent(MyFixedCam, "cam1:")
            image = ADComponent(ImagePlugin, "image1:")
            tiff1 = ADComponent(
                AD_EpicsFileNameTIFFPlugin,
                "TIFF1:",
                write_path_template=WRITE_PATH_TEMPLATE,
                read_path_template=READ_PATH_TEMPLATE,
            )
            pva = ADComponent(PvaPlugin, "Pva1:")

    (new in apstools release 1.6.2)
    """

    pass


class CamMixin_V3_1_1(CamBase):
    """
    Update cam support to AD release 3.1.1.

    (new in release 1.6.3)
    """

    _cam_release = "3.1.1"
    pool_max_buffers = None
    acquire_busy = ADComponent(EpicsSignalRO, "AcquireBusy")
    offset = ADComponent(EpicsSignalWithRBV, "Offset")
    wait_for_plugins = ADComponent(EpicsSignal, "WaitForPlugins")

    @property
    def is_busy(self):
        signal = self.acquire_busy
        return signal.get() in (1, signal.enum_strs[1])


class CamMixin_V34(CamMixin_V3_1_1):
    """
    Update cam support to AD release 3.1.1.

    (new in release 1.6.3)
    """

    _cam_release = "3.4"


class SingleTrigger_V34(SingleTrigger):
    """
    Variation of ophyd's SingleTrigger mixin supporting AcquireBusy.

    (new in release 1.6.3)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        has_v34_cam_features = (
            hasattr(self, "cam")
            and hasattr(self.cam, "_cam_release")
            and version.parse(self.cam._cam_release) >= version.parse("3.4")
        )
        if has_v34_cam_features:
            self._acquisition_busy_signal = self.cam.acquire_busy
        else:
            # backwards compatibility
            self._acquisition_busy_signal = self._acquisition_signal

    def stage(self):
        """Prepare device settings before data acquisition."""
        self._acquisition_busy_signal.subscribe(self._acquire_changed)
        super(SingleTrigger, self).stage()  # from grandparent

    def unstage(self):
        """Restore device settings after data acquisition."""
        super(SingleTrigger, self).unstage()  # from grandparent
        self._acquisition_busy_signal.clear_sub(self._acquire_changed)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Ophyd support for fields common to all EPICS records


Public Structures

.. autosummary::

    ~EpicsRecordDeviceCommonAll
    ~EpicsRecordInputFields
    ~EpicsRecordOutputFields
    ~EpicsRecordFloatFields
    ~EpicsSynAppsRecordEnableMixin

:see: https://wiki-ext.aps.anl.gov/epics/index.php/RRM_3-14_dbCommon
:see: https://wiki-ext.aps.anl.gov/epics/index.php/RRM_3-14_Common
"""

from ophyd.device import Device, Component
from ophyd import EpicsSignal, EpicsSignalRO


class EpicsRecordDeviceCommonAll(Device):
    """
    Many of the fields common to all EPICS records.

    Some fields are not included because they are not interesting to
    an EPICS client or are already provided in other support.
    """

    description = Component(EpicsSignal, ".DESC", kind="config")
    processing_active = Component(EpicsSignalRO, ".PACT", kind="omitted")
    scanning_rate = Component(EpicsSignal, ".SCAN", kind="config")
    disable_value = Component(EpicsSignal, ".DISV", kind="config")
    scan_disable_input_link_value = Component(EpicsSignal, ".DISA", kind="config")
    scan_disable_value_input_link = Component(EpicsSignal, ".SDIS", kind="config")
    process_record = Component(EpicsSignal, ".PROC", kind="omitted", put_complete=True)
    forward_link = Component(EpicsSignal, ".FLNK", kind="config")
    trace_processing = Component(EpicsSignal, ".TPRO", kind="omitted")
    device_type = Component(EpicsSignalRO, ".DTYP", kind="config")

    alarm_status = Component(EpicsSignalRO, ".STAT", kind="config")
    alarm_severity = Component(EpicsSignalRO, ".SEVR", kind="config")
    new_alarm_status = Component(EpicsSignalRO, ".NSTA", kind="config")
    new_alarm_severity = Component(EpicsSignalRO, ".NSEV", kind="config")
    disable_alarm_severity = Component(EpicsSignal, ".DISS", kind="config")


class EpicsRecordInputFields(Device):
    """
    Some fields common to EPICS input records.
    """

    input_link = Component(EpicsSignal, ".INP", kind="config")
    raw_value = Component(EpicsSignal, ".RVAL", kind="config")
    final_value = Component(EpicsSignal, ".VAL", kind="normal")

    # will ignore simulation mode fields

    @property
    def value(self):
        return self.final_value.get()


class EpicsRecordOutputFields(Device):
    """
    Some fields common to EPICS output records.
    """

    output_link = Component(EpicsSignal, ".OUT", kind="config")
    raw_value = Component(EpicsSignal, ".RVAL", kind="config")
    output_value = Component(EpicsSignal, ".OVAL", kind="normal")
    readback_value = Component(EpicsSignalRO, ".RBV", kind="hinted")
    desired_output_location = Component(EpicsSignal, ".DOL", kind="config")
    output_mode_select = Component(EpicsSignal, ".OMSL", kind="config")
    desired_value = Component(EpicsSignal, ".VAL", kind="normal")

    # will ignore simulation mode fields

    @property
    def value(self):
        return self.desired_value.get()


class EpicsRecordFloatFields(Device):
    """
    Some fields common to EPICS records supporting floating point values.
    """

    units = Component(EpicsSignal, ".EGU", kind="config")
    precision = Component(EpicsSignal, ".PREC", kind="config")

    monitor_deadband = Component(EpicsSignal, ".MDEL", kind="config")

    # upper and lower display limits for the VAL, CVAL, HIHI, HIGH, LOW, and LOLO fields
    high_operating_range = Component(EpicsSignal, ".HOPR", kind="config")
    low_operating_range = Component(EpicsSignal, ".LOPR", kind="config")

    hihi_alarm_limit = Component(EpicsSignal, ".HIHI", kind="config")
    high_alarm_limit = Component(EpicsSignal, ".HIGH", kind="config")
    low_alarm_limit = Component(EpicsSignal, ".LOW", kind="config")
    lolo_alarm_limit = Component(EpicsSignal, ".LOLO", kind="config")
    hihi_alarm_severity = Component(EpicsSignal, ".HHSV", kind="config")
    high_alarm_severity = Component(EpicsSignal, ".HSV", kind="config")
    low_alarm_severity = Component(EpicsSignal, ".LSV", kind="config")
    lolo_alarm_severity = Component(EpicsSignal, ".LLSV", kind="config")
    alarm_hysteresis = Component(EpicsSignal, ".HYST", kind="config")


class EpicsSynAppsRecordEnableMixin(Device):
    """Supports ``{PV}Enable`` feature from user databases."""
    enable = Component(EpicsSignal, "Enable", kind="config")

    def reset(self):
        """set all fields to default values"""
        self.enable.put(self.enable.enum_strs[1])  # Enable
        super().reset()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Ophyd support for the EPICS sub record

https://epics.anl.gov/base/R7-0/6-docs/subRecord.html

Public Structures

.. autosummary::

    ~UserAverageN
    ~UserAverageDevice
    ~SubRecord
    ~SubRecordChannel
"""

from collections import OrderedDict
from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as DDC
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import FormattedComponent as FC

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsRecordFloatFields
from .. import utils as APS_utils


CHANNEL_LETTERS_LIST = "A B C D E F G H I J K L".split()


class SubRecordChannel(Device):
    """
    Number channel of a sub record: A-L

    .. index:: Ophyd Device; synApps SubRecordChannel

    .. autosummary::

        ~reset
    """

    input_value = FC(EpicsSignal, "{prefix}.{_ch}", kind="config")
    input_pv = FC(EpicsSignal, "{prefix}.INP{_ch}", kind="config", string=True)

    def __init__(self, prefix, letter, **kwargs):
        self._ch = letter
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.input_pv.put("")
        self.input_value.put(0)


def _channels(input_list):
    defn = OrderedDict()
    for nsym in input_list:
        defn[nsym] = (SubRecordChannel, "", {"letter": nsym})

    return defn


class SubRecord(EpicsRecordFloatFields, EpicsRecordDeviceCommonAll):
    """
    EPICS base sub record support in ophyd

    .. index:: Ophyd Device; SubRecord

    .. autosummary::

        ~reset

    :see: http://htmlpreview.github.io/?https://github.com/epics-modules/calc/blob/R3-6-1/documentation/calcDocs.html
    """
    units = Cpt(EpicsSignal, ".EGU", kind="config")
    precision = Cpt(EpicsSignal, ".PREC", kind="config")

    calculated_value = Cpt(EpicsSignal, ".VAL", kind="hinted")
    initroutine = Cpt(EpicsSignal, ".INAM", kind="config", string=True)
    subroutine = Cpt(EpicsSignal, ".SNAM", kind="config", string=True)

    channels = DDC(_channels(CHANNEL_LETTERS_LIST))

    read_attrs = APS_utils.itemizer("channels.%s", CHANNEL_LETTERS_LIST)
    hints = {"fields": read_attrs}

    @property
    def value(self):
        return self.calculated_value.get()

    def reset(self):
        """set all fields to default values"""
        pvname = self.description.pvname.split(".")[0]
        self.scanning_rate.put("Passive")
        self.description.put(pvname)
        self.units.put("")
        self.precision.put("5")

        self.initroutine.put("")  # initSubAve in userAve
        self.subroutine.put("")  # SubAve in userAve
        self.calculated_value.put(0)

        self.forward_link.put("")

        for letter in self.channels.read_attrs:
            channel = getattr(self.channels, letter)
            if isinstance(channel, SubRecordChannel):
                channel.reset()
        self.read_attrs = ["channels.%s" % c for c in CHANNEL_LETTERS_LIST]
        self.hints = {"fields": self.read_attrs}


class UserAverageN(EpicsRecordFloatFields, EpicsRecordDeviceCommonAll):
    """
    EPICS synApps XXX IOC setup of user average: ``$(P):userAve$(N)``

    This database uses a sub record for most features plus additional
    records to support done, acquire, clear, and other features.

    It *uses* a ``sub`` record and other, hence not exactly a :meth:`SubRecord`.
    """
    enable = Cpt(EpicsSignal, "Enable", kind="omitted", string=True, put_complete=True)

    averaged_value = Cpt(EpicsSignalRO, ".VAL", kind="hinted")
    channel = Cpt(EpicsSignal, ".INPB", kind="config", string=True, put_complete=True)

    initroutine = Cpt(EpicsSignalRO, ".INAM", kind="config", string=True)
    subroutine = Cpt(EpicsSignalRO, ".SNAM", kind="config", string=True)

    number_samples = Cpt(EpicsSignal, ".A", kind="config", put_complete=True)
    input_value = Cpt(EpicsSignalRO, ".B", kind="normal")
    current_sample = Cpt(EpicsSignalRO, ".E", kind="config")
    average = Cpt(EpicsSignalRO, ".H", kind="hinted")
    line_fit = Cpt(EpicsSignalRO, ".I", kind="hinted")
    auto = Cpt(EpicsSignalRO, ".J", kind="hinted")
    slope = Cpt(EpicsSignalRO, ".K", kind="hinted")
    corr_coef = Cpt(EpicsSignalRO, ".L", kind="hinted")

    mode = Cpt(EpicsSignal, "_mode", kind="config", string=True, put_complete=True)
    algorithm = Cpt(EpicsSignal, "_algorithm", kind="config", string=True, put_complete=True)

    clear = Cpt(EpicsSignal, ".C", kind="omitted", put_complete=True)
    acquire = Cpt(EpicsSignal, "_acquire", kind="omitted", put_complete=True)
    busy = Cpt(EpicsSignal, "_busy", kind="omitted")

    def reset(self):
        """set all fields to default values"""
        self.enable.put(self.enable.enum_strs[1])  # enable
        pvname = self.description.pvname.split(".")[0]
        self.description.put(pvname)
        for attr in "scanning_rate mode algorithm".split():
            obj = getattr(self, attr)
            obj.put(obj.enum_strs[0])
        self.channel.put("")
        self.number_samples.put(0)
        self.precision.put(0)


class UserAverageDevice(Device):
    """
    EPICS synApps XXX IOC setup of user averaging sub records: ``$(P):userAve$(N)``

    .. index:: Ophyd Device; synApps UserAverageDevice

    .. autosummary::

        ~reset

    """

    enable = Cpt(EpicsSignal, "userAveEnable", kind="omitted")
    average1 = Cpt(UserAverageN, "userAve1")
    average2 = Cpt(UserAverageN, "userAve2")
    average3 = Cpt(UserAverageN, "userAve3")
    average4 = Cpt(UserAverageN, "userAve4")
    average5 = Cpt(UserAverageN, "userAve5")
    average6 = Cpt(UserAverageN, "userAve6")
    average7 = Cpt(UserAverageN, "userAve7")
    average8 = Cpt(UserAverageN, "userAve8")
    average9 = Cpt(UserAverageN, "userAve9")
    average10 = Cpt(UserAverageN, "userAve10")

    def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        self.average1.reset()
        self.average2.reset()
        self.average3.reset()
        self.average4.reset()
        self.average5.reset()
        self.average6.reset()
        self.average7.reset()
        self.average8.reset()
        self.average9.reset()
        self.average10.reset()
        self.read_attrs = ["average%d" % (c + 1) for c in range(10)]
        self.read_attrs.insert(0, "enable")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



import pytest

from ..calcout import CalcoutRecord
from ..sub import SubRecord
from ..sub import UserAverageN
from ..sub import UserAverageDevice
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing


@pytest.fixture(scope="function")
def ave():
    ave = UserAverageN(f"{IOC}userAve9", name="ave")
    ave.wait_for_connection()
    ave.enable.put("E")
    yield ave
    ave.reset()


@pytest.fixture(scope="function")
def calc():
    calc = CalcoutRecord(f"{IOC}userCalcOut9", name="calc")
    calc.wait_for_connection()
    calc.enable.put("E")
    yield calc
    calc.reset()


# @pytest.fixture(scope="function")
# def sub():
#     # Is there some _other_ sub record to use for testing?
#     sub = SubRecord(f"{IOC}userAve9", name="sub")
#     sub.wait_for_connection()
#     sub.enable.put("E")
#     yield sub
#     sub.reset()


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [SubRecord, f"{IOC}userAve10", False, "read_attrs", 12],
        [SubRecord, f"{IOC}userAve10", False, "configuration_attrs", 65],
        [SubRecord, f"{IOC}userAve10", True, "read()", 1],
        [SubRecord, f"{IOC}userAve10", True, "summary()", 148],

        [UserAverageN, f"{IOC}userAve10", False, "read_attrs", 7],
        [UserAverageN, f"{IOC}userAve10", False, "configuration_attrs", 33],
        [UserAverageN, f"{IOC}userAve10", True, "read()", 7],
        [UserAverageN, f"{IOC}userAve10", True, "summary()", 101],

        [UserAverageDevice, IOC, False, "read_attrs", 80],
        [UserAverageDevice, IOC, False, "configuration_attrs", 340],
        [UserAverageDevice, IOC, True, "read()", 70],
        [UserAverageDevice, IOC, True, "summary()", 835],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


# def test_sub_reset(sub):
#     assert isinstance(sub, SubRecord)
#     # TODO: need a sub record that can be reset, no general purpose available


def test_useraverage_reset(ave):
    pvname = ave.prefix.split(".")[0]

    assert ave.initroutine.get() == "initSubAve"
    assert ave.subroutine.get() == "SubAve"

    ave.number_samples.put(10)
    ave.channel.put(f"{pvname}.VAL")
    ave.precision.put(3)
    ave.mode.put("ONE-SHOT")
    ave.algorithm.put("FIT-LINE")

    ave.reset()
    short_delay_for_EPICS_IOC_database_processing()
    assert ave.initroutine.get() == "initSubAve"
    assert ave.subroutine.get() == "SubAve"

    assert ave.description.get() == pvname
    assert ave.enable.get() == "E"
    assert ave.channel.get() == ""
    # TODO: comes back as 1 : Why is that?
    # assert ave.number_samples.get() == 0
    assert ave.precision.get() == 0
    for attr in "mode algorithm scanning_rate".split():
        obj = getattr(ave, attr)
        assert obj.get() in (0, obj.enum_strs[0]), attr

# FIXME: 2022-01-20: skipped per #627
# def test_useraveragedevice_random(ave, calc):
#     # setup a random number generator
#     calc.calculation.put("RNDM")

#     # average the RNG
#     nsamples = 100
#     ave.channel.put(calc.prefix)
#     ave.number_samples.put(nsamples)
#     ave.precision.put(3)
#     ave.mode.put("ONE-SHOT")
#     ave.acquire.put(1)
#     for i in range(nsamples):
#         calc.process_record.put(1)
#         ave.process_record.put(1)
#         short_delay_for_EPICS_IOC_database_processing()
#         if ave.current_sample.get() == ave.number_samples.get():
#             break
#         assert ave.busy.get(as_string=False) == 1, i
#     assert ave.busy.get(as_string=False) == 0
#     # test that average = 0.5 as n -> inf
#     assert abs(ave.averaged_value.get() - 0.5) <= 0.1



"""
OverrideParameters
+++++++++++++++++++++++++++++++++++++++

Define parameters that can be overridden from a user configuration file.

EXAMPLE:

Create an ``overrides`` object in a new file ``override_params.py``::

    import apstools.utils
    overrides = apstools.utils.OverrideParameters()

When code supports a parameter for which a user can provide
a local override, the code should import the ``overrides``
object (from the ``override_params`` module),
and then register the parameter name, such as this example::

    from override_params import overrides
    overrides.register("minimum_step")

Then later::

    minstep = overrides.pick("minimum_step", 45e-6)

In the user's configuration file that will override
the value of ``45e-6`` (such as can be loaded via
``%run -i user.py``), import the `overrides``
object (from the ``override_params`` module)::

    from override_params import overrides

and then override the attribute(s) as desired::

    overrides.set("minimum_step", 1.0e-5)

With this override in place, the ``minstep`` value
(from :meth:`~apstools.utils.OverrideParameters.pick()`)
will be ``1e-5``.

Get a pandas DataFrame object with all the overrides::

    overrides.summary()

which returns this table::

        parameter    value
    0  minimum_step  0.00001

----

.. autosummary::

   ~OverrideParameters
"""


import pandas as pd


class OverrideParameters:
    """
    Define parameters that can be overridden from a user configuration file.

    NOTE: This is a pure Python object, not using ophyd.

    .. autosummary::

       ~pick
       ~register
       ~reset
       ~reset_all
       ~set
       ~summary

    (new in apstools 1.5.2)
    """

    def __init__(self):
        # ALWAYS use ``overrides.undefined`` for comparisons and resets.
        self.undefined = object()
        self._parameters = {}

    def register(self, parameter_name):
        """
        Register a new parameter name to be supported by user overrides.
        """
        if parameter_name not in self._parameters:
            self._parameters[parameter_name] = self.undefined

    def _check_known(self, parameter_name):
        if parameter_name not in self._parameters:
            raise KeyError(
                f"Unknown parameter {parameter_name}.  First call register()."
            )

    def set(self, parameter_name, value):
        """
        Define an override value for a known parameter.
        """
        self._check_known(parameter_name)
        self._parameters[parameter_name] = value

    def reset(self, parameter_name):
        """
        Remove an override value for a known parameter.  (sets it to undefined)
        """
        self._check_known(parameter_name)
        self._parameters[parameter_name] = self.undefined

    def reset_all(self):
        """
        Remove override values for all known parameters. (sets all to undefined)
        """
        for parm in self._parameters.keys():
            self.reset(parm)

    def pick(self, parameter, default):
        """
        Return either the override parameter value if defined, or the default.
        """
        value = self._parameters.get(parameter, default)
        if value == self.undefined:
            value = default
        return value

    def summary(self):
        """
        Return a pandas DataFrame with all overrides.

        Parameter names that have no override value will be reported
        as ``--undefined--``.
        """
        parameters = []
        values = []
        for parm in sorted(self._parameters.keys()):
            parameters.append(parm)
            values.append(self.pick(parm, "--undefined--"))
        return pd.DataFrame(dict(parameter=parameters, value=values))

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Ophyd support for the EPICS synApps luascript record

EXAMPLES::

    import apstools.synApps
    scripts = apstools.synApps.UserScriptsDevice("xxx:", name="scripts")
    scripts.reset()

.. autosummary::

    ~UserScriptsDevice
    ~LuascriptRecord
    ~LuascriptRecordNumberInput
    ~LuascriptRecordStringInput

:see: https://epics-lua.readthedocs.io/en/latest/luascriptRecord.html
"""


from collections import OrderedDict
from ophyd import Device
from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as DDC
from ophyd import FormattedComponent as FC
from ophyd import EpicsSignal

from ._common import EpicsRecordDeviceCommonAll
from .. import utils as APS_utils


INPUT_LETTERS_LIST = "A B C D E F G H I J".split()


class _LuascriptRecordInputBase(Device):
    """
    (internal): base class for one input of a LuascriptRecord
    """

    # Components must be defined in subclass

    # read_attrs = [
    #     "input_value",
    # ]
    # hints = {"fields": read_attrs}

    def __init__(self, prefix, letter, **kwargs):
        self._ch = letter
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        raise NotImplementedError("Must define reset() method in subclass.")


class LuascriptRecordNumberInput(_LuascriptRecordInputBase):
    """
    number input of a synApps luascript record: A-J

    .. index:: Ophyd Device; synApps LuascriptRecordNumberInput
    """

    pv_link = FC(EpicsSignal, "{prefix}.INP{_ch}", kind="config", string=True)
    input_value = FC(EpicsSignal, "{prefix}.{_ch}", kind="config")
    description = FC(EpicsSignal, "{prefix}.{_ch}DSC", kind="config", string=True)

    def reset(self):
        """set all fields to default values"""
        self.pv_link.put("")
        self.input_value.put(0)
        self.description.put("")


class LuascriptRecordStringInput(_LuascriptRecordInputBase):
    """
    string input of a synApps luascript record: AA-JJ

    .. index:: Ophyd Device; synApps LuascriptRecordStringInput
    """

    pv_link = FC(EpicsSignal, "{prefix}.IN{_ch}", kind="config", string=True)
    input_value = FC(EpicsSignal, "{prefix}.{_ch}", kind="config", string=True)
    description = FC(EpicsSignal, "{prefix}.{_ch}DN", kind="config", string=True)

    def reset(self):
        """set all fields to default values"""
        self.pv_link.put("")
        self.input_value.put("")
        self.description.put("")


def _inputs(input_list):
    defn = OrderedDict()
    for nsym in input_list:
        defn[nsym] = (LuascriptRecordNumberInput, "", {"letter": nsym})
        ssym = nsym + nsym
        defn[ssym] = (LuascriptRecordStringInput, "", {"letter": ssym})

    return defn


class LuascriptRecord(EpicsRecordDeviceCommonAll):
    """
    EPICS synApps luascript record: used as ``$(P):userScript$(N)``

    .. index:: Ophyd Device; synApps LuascriptRecord

    .. autosummary::

        ~reset

    """

    precision = Cpt(EpicsSignal, ".PREC", kind="config")

    number_value = Cpt(EpicsSignal, ".VAL", kind="hinted")
    string_value = Cpt(EpicsSignal, ".SVAL", kind="hinted", string=True)
    code = Cpt(EpicsSignal, ".CODE", kind="config", string=True)

    output_link = Cpt(EpicsSignal, ".OUT", kind="config")
    reload_script = Cpt(EpicsSignal, ".RELO", kind="omitted", put_complete=True)
    output_execute_option = Cpt(EpicsSignal, ".OOPT", kind="config")

    error_message = Cpt(EpicsSignal, ".ERR", kind="config", string=True)

    read_attrs = APS_utils.itemizer("inputs.%s", INPUT_LETTERS_LIST)
    hints = {"fields": read_attrs}

    inputs = DDC(_inputs(INPUT_LETTERS_LIST))

    def reset(self):
        """set all fields to default values"""
        pvname = self.description.pvname.split(".")[0]
        self.description.put(pvname)
        self.scanning_rate.put("Passive")
        self.code.put("")
        self.precision.put("5")
        self.output_link.put("")
        self.number_value.put(0)
        self.string_value.put("")
        new_read_attrs = []
        for letter in self.inputs.read_attrs:
            if "." not in letter:
                input = getattr(self.inputs, letter)
                if isinstance(input, (LuascriptRecordNumberInput, LuascriptRecordStringInput)):
                    new_read_attrs.append(f"inputs.{letter}")
                    input.reset()

        self.read_attrs = new_read_attrs
        self.hints = {"fields": new_read_attrs}

        self.read_attrs.append("number_value")
        self.read_attrs.append("string_value")


class UserScriptsDevice(Device):
    """
    EPICS synApps XXX IOC setup of user lua scripts: ``$(P):userScript$(N)``

    .. index:: Ophyd Device; synApps UserScriptsDevice

    .. autosummary::

        ~reset

    """

    enable = Cpt(EpicsSignal, "userScriptEnable", kind="config")
    script0 = Cpt(LuascriptRecord, "userScript0")
    script1 = Cpt(LuascriptRecord, "userScript1")
    script2 = Cpt(LuascriptRecord, "userScript2")
    script3 = Cpt(LuascriptRecord, "userScript3")
    script4 = Cpt(LuascriptRecord, "userScript4")
    script5 = Cpt(LuascriptRecord, "userScript5")
    script6 = Cpt(LuascriptRecord, "userScript6")
    script7 = Cpt(LuascriptRecord, "userScript7")
    script8 = Cpt(LuascriptRecord, "userScript8")
    script9 = Cpt(LuascriptRecord, "userScript9")

    def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        enable = self.enable.get()
        self.enable.put("Enable")
        self.script0.reset()
        self.script1.reset()
        self.script2.reset()
        self.script3.reset()
        self.script4.reset()
        self.script5.reset()
        self.script6.reset()
        self.script7.reset()
        self.script8.reset()
        self.script9.reset()
        self.enable.put(enable)
        self.read_attrs = [f"script{c}" for c in range(10)]

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Device information
++++++++++++++++++

.. autosummary::

   ~listdevice
"""


from collections import defaultdict
from ophyd import Device
from ophyd import Signal
from ophyd.signal import EpicsSignalBase

import datetime
import logging
import pandas as pd


logger = logging.getLogger(__name__)
pd.set_option("display.max_rows", None)


def _all_signals(base):
    if isinstance(base, Signal):
        return [base]
    items = []
    if hasattr(base, "component_names"):
        for k in base.component_names:
            obj = getattr(base, k)
            if isinstance(obj, (Device, Signal)):
                items += _all_signals(obj)
    return items


def _get_named_child(obj, nm):
    """
    return named child of ``obj`` or None
    """
    try:
        child = getattr(obj, nm)
        return child
    except TimeoutError:
        logger.debug(f"timeout: {obj.name}_{nm}")
        return "TIMEOUT"


def _get_pv(obj):
    """
    Return PV name, prefix, or None from ophyd object.
    """
    if hasattr(obj, "pvname"):
        return obj.pvname
    elif hasattr(obj, "prefix"):
        return obj.prefix


def _list_epics_signals(obj):
    """
    Return a list of the EPICS signals in obj.

    RETURNS

    list of ophyd objects that are children of ``obj``
    """
    # import pdb; pdb.set_trace()
    if isinstance(obj, EpicsSignalBase):
        return [obj]
    elif isinstance(obj, Device):
        items = []
        for nm in obj.component_names:
            child = _get_named_child(obj, nm)
            if child in (None, "TIMEOUT"):
                continue
            result = _list_epics_signals(child)
            if result is not None:
                items.extend(result)
        return items


def listdevice(
    obj,
    scope=None,
    cname=False,
    dname=True,
    show_pv=False,
    use_datetime=True,
    show_ancient=True,
):
    """
    Describe the signal information from device ``obj`` in a pandas DataFrame.

    Look through all subcomponents to find all the signals to be shown.

    PARAMETERS

    obj
        *object* : Instance of ophyd Signal or Device.
    scope
        *str* or None : Scope of content to be shown.

        - ``"full"`` (or ``None``) shows all Signal components
        - ``"epics"`` shows only EPICS-based Signals
        - ``"read"`` shows only the signals returned by ``obj.read()``

        default: ``None``
    cname
        *bool* : Show the _control_ (Python, dotted) name in column ``name``.

        default: ``False``
    dname
        *bool* : Show the _data_ (databroker, with underlines) name in
        column ``data name``.

        default: ``True``
    show_pv
        *bool* : Show the EPICS process variable (PV) name in
        column ``PV``.

        default: ``False``
    use_datetime
        *bool* : Show the EPICS timestamp (time of last update) in
        column ``timestamp``.

        default: ``True``
    show_ancient
        *bool* : Show uninitialized EPICS process variables.

        In EPICS, an uninitialized PV has a timestamp of 1990-01-01 UTC.
        This option enables or suppresses ancient values identified
        by timestamp from 1989.  These are values only defined in
        the original ``.db`` file.

        default: ``True``
    """
    scope = (scope or "full").lower()
    signals = _all_signals(obj)
    if scope in ("full", "epics"):
        if scope == "epics":
            signals = [s for s in signals if isinstance(s, EpicsSignalBase)]
    elif scope == "read":
        reading = obj.read()
        signals = [s for s in signals if s.name in reading]
    else:
        # fmt: off
        raise KeyError(
            f"Unknown scope='{scope}'."
            " Must be one of None, 'full', 'epics', 'read'"
        )
        # fmt: on

    # in EPICS, an uninitialized PV has a timestamp of 1990-01-01 UTC
    UNINITIALIZED = datetime.datetime.timestamp(
        datetime.datetime.fromisoformat("1990-06-01")
    )

    if not cname and not dname:
        cname = True

    dd = defaultdict(list)
    for signal in signals:
        if scope != "epics" or isinstance(signal, EpicsSignalBase):
            ts = getattr(signal, "timestamp", 0)
            if show_ancient or (ts >= UNINITIALIZED):
                if cname:
                    head = obj
                    while head.dotted_name != "":
                        # walk back to the head
                        head = head.parent
                    dd["name"].append(f"{head.name}.{signal.dotted_name}")
                if dname:
                    dd["data name"].append(signal.name)
                if show_pv:
                    dd["PV"].append(_get_pv(signal) or "")
                dd["value"].append(signal.get())
                if use_datetime:
                    dd["timestamp"].append(datetime.datetime.fromtimestamp(ts))

    return pd.DataFrame(dd)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from ..input_plan import request_input
from bluesky import RunEngine
import databroker
import pytest


cat = databroker.temp()
RE = RunEngine({})
RE.subscribe(cat.v1.insert)


def my_plan(msg, default, agree, bypass):
    yield from request_input(
        msg=msg,
        default=default,
        agree=agree,
        bypass=bypass,
    )


@pytest.mark.parametrize(
    "msg, default, agree, bypass",
    [
        # TODO: re-enable pending bluesky/bluesky#1550
        # ["Continue?", "y", "y", False],
        ["Continue?", "y", "y", True],
        [None, None, None, True],
        ["Continue?", "n", "y", True],
        ["Continue?", "n", ["y", "Sure", "certainly"], True],
        ["Continue?", "n", "n", True],
    ],
)
def test(msg, default, agree, bypass):
    if bypass:
        uids = RE(my_plan(msg, default, agree, bypass))
        assert isinstance(uids, (list, tuple))
    else:
        with pytest.raises(ValueError) as exinfo:
            RE(my_plan(msg, default, agree, bypass))
        assert str(exinfo.value).startswith("Invalid file object")



"""
Ophyd support for the EPICS sseq (string sequence) record


Public Structures

.. autosummary::

    ~EditStringSequence
    ~SseqRecord
    ~UserStringSequenceDevice
    ~UserStringSequenceN
"""

from collections import OrderedDict
from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as DDC
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import FormattedComponent as FC

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsSynAppsRecordEnableMixin


STEP_LIST = [f"step{i+1}" for i in range(10)]  # step1, step2, step10


class sseqRecordStep(Device):
    """
    EPICS synApps sseq single step of an sseq record.

    Step of a synApps sseq record: 1..10  (note: for 10, the PVs use "A")

    .. index:: Ophyd Device; synApps sseqRecordStep

    .. autosummary::

        ~reset
    """

    input_pv = FC(EpicsSignal, "{self.prefix}.DOL{self._step}", kind="config")
    input_pv_valid = FC(EpicsSignalRO, "{self.prefix}.DOL{self._step}V", kind="config")
    delay = FC(EpicsSignal, "{self.prefix}.DLY{self._step}", kind="config")
    string_value = FC(
        EpicsSignal, "{self.prefix}.STR{self._step}", string=True, kind="hinted"
    )
    numeric_value = FC(EpicsSignal, "{self.prefix}.DO{self._step}", kind="hinted")
    output_pv = FC(EpicsSignal, "{self.prefix}.LNK{self._step}", kind="config")
    output_pv_valid = FC(EpicsSignalRO, "{self.prefix}.LNK{self._step}V", kind="config")

    waiting_completion = FC(
        EpicsSignalRO, "{self.prefix}.WTG{self._step}", kind="config"
    )
    wait_completion = FC(EpicsSignal, "{self.prefix}.WAIT{self._step}", kind="config")
    wait_error = FC(EpicsSignalRO, "{self.prefix}.WERR{self._step}", kind="config")

    def __init__(self, prefix, step, **kwargs):
        names = "_123456789A"  # step #10 (1-based) is called "A"
        self._step = names[step]
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.input_pv.put("")
        self.delay.put(0)
        self.numeric_value.put(0)  # EPICS will set string_value from this
        self.output_pv.put("")
        self.wait_completion.put("NoWait")


def _steps(step_list):
    defn = OrderedDict()
    for step in step_list:
        step_number = int(step[4:])
        defn[step] = (sseqRecordStep, "", {"step": step_number})
    return defn


class SseqRecord(EpicsRecordDeviceCommonAll):
    """
    EPICS synApps sseq record support in ophyd

    .. index:: Ophyd Device; synApps SseqRecord

    .. autosummary::

        ~abort
        ~reset

    :see: https://htmlpreview.github.io/?https://raw.githubusercontent.com/epics-modules/calc/R3-6-1/documentation/sseqRecord.html
    """

    precision = Cpt(EpicsSignal, ".PREC", kind="config")
    busy = Cpt(EpicsSignalRO, ".PREC", kind="config")
    _abort = Cpt(EpicsSignal, ".ABORT", kind="omitted")

    selection_link = Cpt(EpicsSignal, ".SELL", kind="config")
    selection_mask = Cpt(EpicsSignal, ".SELM", kind="config")
    selection_number = Cpt(EpicsSignal, ".SELN", kind="config")

    steps = DDC(_steps(STEP_LIST))

    def abort(self):
        """
        .ABORT is a push button.  Send a 1 to the PV to "push" it.

        Push this button without a timeout from the .put() method.
        """
        self._abort.put(1, use_complete=False, force=True)

    def reset(self):
        """set all fields to default values"""
        self.scanning_rate.put("Passive")
        self.description.put(self.description.pvname.split(".")[0])
        self.forward_link.put("")
        self.precision.put(5)
        self.selection_link.put("")
        self.selection_mask.put("All")
        self.selection_number.put(1)
        for ch in self.steps.component_names:
            step = getattr(self.steps, ch)
            if isinstance(step, sseqRecordStep):
                step.reset()
        self.hints["fields"] = ["steps_%s" % c for c in self.steps.component_names]
        self.read_attrs = ["steps.%s" % c for c in STEP_LIST]


class UserStringSequenceN(EpicsSynAppsRecordEnableMixin, SseqRecord):
    """Single instance of the userStringSeqN database."""


class UserStringSequenceDevice(Device):
    """
    EPICS synApps XXX IOC setup of userStringSeqs: ``$(P):userStringSeq$(N)``

    Note: This will connect more than 1,000 EpicsSignal objects!

    .. index:: Ophyd Device; synApps UserStringSequenceDevice

    .. autosummary::

        ~reset
    """

    enable = Cpt(EpicsSignal, "userStringSeqEnable", kind="config")
    sseq1 = Cpt(UserStringSequenceN, "userStringSeq1")
    sseq2 = Cpt(UserStringSequenceN, "userStringSeq2")
    sseq3 = Cpt(UserStringSequenceN, "userStringSeq3")
    sseq4 = Cpt(UserStringSequenceN, "userStringSeq4")
    sseq5 = Cpt(UserStringSequenceN, "userStringSeq5")
    sseq6 = Cpt(UserStringSequenceN, "userStringSeq6")
    sseq7 = Cpt(UserStringSequenceN, "userStringSeq7")
    sseq8 = Cpt(UserStringSequenceN, "userStringSeq8")
    sseq9 = Cpt(UserStringSequenceN, "userStringSeq9")
    sseq10 = Cpt(UserStringSequenceN, "userStringSeq10")

    def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        for c in self.component_names:
            if not c.startswith("sseq"):
                continue
            getattr(self, c).reset()
        self.read_attrs = self.component_names


class EditStringSequence(Device):
    """
    EPICS synApps sseq support to quickly re-arrange steps.

    See the ``editSseq_more`` GUI screen for assistance.
    """
    record_name = Cpt(EpicsSignal, "ES:recordName", kind="config")
    command = Cpt(EpicsSignal, "ES:command", kind="config")
    message_acknowledge = Cpt(EpicsSignal, "ES:OperAck", kind="config")
    message = Cpt(EpicsSignalRO, "ES:message", kind="normal")
    alert = Cpt(EpicsSignalRO, "ES:Alert", kind="normal")
    debug = Cpt(EpicsSignal, "ES:Debug", kind="config")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



import databroker
import datetime
import intake
import pandas as pd
import pytest

from .. import getDefaultNamespace
from ... import utils as utils


TEST_CATALOG_NAME = "apstools_test"


@pytest.fixture(scope="function")
def cat():
    cat = databroker.catalog[TEST_CATALOG_NAME]
    return cat


@pytest.fixture(scope="function")
def lr():
    lr = utils.ListRuns()
    lr.cat = databroker.catalog[TEST_CATALOG_NAME]
    lr._check_keys()
    assert len(lr.cat) == 53
    return lr


def test_getDefaultCatalog_none_found():
    with pytest.raises(ValueError) as exinfo:
        utils.getDefaultCatalog()
    assert "Multiple catalog configurations available." in str(exinfo.value)


def test_getDefaultCatalog(cat):
    # put the catalog in the namespace of the called function
    ns = getDefaultNamespace()
    ns.update(dict(cat1=cat))

    cat = utils.getDefaultCatalog()
    assert cat is not None
    assert cat.name == TEST_CATALOG_NAME


def test_getDefaultCatalog_many_found(cat):
    ns = getDefaultNamespace()
    ns.update(dict(cat1=cat, cat2=cat, cat3=cat))

    with pytest.raises(ValueError) as exinfo:
        utils.getDefaultCatalog()
    assert "Multiple catalog objects available." in str(exinfo.value)


def test_getCatalog():
    # get by name of configuration YAML file
    ret = utils.getCatalog(TEST_CATALOG_NAME)
    assert ret.name == TEST_CATALOG_NAME

    # get by supplying the catalog
    ret = utils.getCatalog(ret)
    assert ret.name == TEST_CATALOG_NAME

    # fail with configuration YAML file name not found
    bad_name = "no such catalog configuration"
    with pytest.raises(KeyError) as exinfo:
        utils.getCatalog(bad_name)
    assert bad_name in str(exinfo.value)


def test_getDefaultNamespace():
    ret = utils.getDefaultNamespace()
    assert ret is not None
    assert isinstance(ret, dict)
    assert len(ret) > 0
    assert "getDefaultNamespace" in ret


def test_findCatalogsInNamespace(cat):
    ns = getDefaultNamespace()
    assert isinstance(ns, dict)

    ns.update(dict(cat1=cat, cat2=cat, cat3=cat))

    cats = utils.findCatalogsInNamespace()
    assert isinstance(cat, intake.Catalog)
    assert len(cats) == 3


def test_ListRuns(cat):
    lr = utils.ListRuns()
    assert lr is not None
    assert lr.cat is None
    assert lr.keys is None
    assert lr.missing == ""
    assert lr.num == 20
    assert lr.query is None
    assert lr.reverse is True
    assert lr.since is None
    assert lr.until is None
    assert lr.sortby == "time"
    assert lr.timefmt == "%Y-%m-%d %H:%M:%S"

    lr.cat = cat
    lr._check_cat()
    assert cat == lr.cat  # expect unchanged


def test_ListRuns_keys(lr):
    # use default keys
    lr.keys = None
    lr._check_keys()
    assert lr.keys is not None
    keys = "scan_id time plan_name detectors".split()  # expected
    assert len(lr.keys) == len(keys)
    for idx, label in enumerate(keys):
        assert lr.keys[idx] == label

    # custom keys as str
    keys = "scan_id plan_name"
    lr.keys = keys
    lr._check_keys()
    keys = keys.split()  # expected
    assert len(lr.keys) == len(keys)
    for idx, label in enumerate(keys):
        assert lr.keys[idx] == label

    # custom keys as [str]
    keys = "plan_name detectors motors".split()
    lr.keys = keys
    lr._check_keys()
    assert len(lr.keys) == len(keys)
    for idx, label in enumerate(keys):
        assert lr.keys[idx] == label

    # check that keys are in results dictionary
    dd = lr.parse_runs()
    assert len(dd) == len(lr.keys)
    for label in lr.keys:
        assert label in dd

    lr.keys.append("exit_status")
    lr.num = 1_000
    dd = lr.parse_runs()
    possibilities = "success fail abort".split()
    possibilities.append("")  # means "no stop document was recorded"
    # possibilities.append("unknown")
    for v in dd["exit_status"]:
        assert v in possibilities

    keys = "scan_id start.time stop.time"
    lr.keys = keys
    lr._check_keys()
    keys = keys.split()  # expected
    assert len(lr.keys) == len(keys)
    for idx, label in enumerate(keys):
        assert lr.keys[idx] == label


def test_ListRuns_missing(lr):
    lr.keys = "scan_id time".split()
    key = "no such key"
    lr.keys.append(key)
    missing = ""  # the default
    dd = lr.parse_runs()
    assert key in dd
    for v in dd[key]:
        assert v == missing

    missing = object()  # a very unexpected representation
    lr.missing = missing
    dd = lr.parse_runs()
    assert key in dd
    for v in dd[key]:
        assert v == missing


def test_ListRuns_to_num(lr):
    # depends on test database
    dd = lr.parse_runs()
    assert 0 <= len(dd["time"]) <= lr.num
    # assert len(dd["time"]) == -1

    lr.num = lr.num // 2
    dd = lr.parse_runs()
    assert 0 <= len(dd["time"]) <= lr.num

    lr.num *= 5
    dd = lr.parse_runs()
    assert 0 <= len(dd["time"]) <= lr.num


def test_ListRuns_query_count(lr):
    lr.query = dict(plan_name="count")
    dd = lr.parse_runs()
    assert 0 <= len(dd["time"]) <= lr.num
    for v in dd["plan_name"]:
        assert v == "count"


# fmt: off
@pytest.mark.parametrize(
    "nruns, query",
    [
        (27, dict(scan_id={"$lt": 20})),
        (26, dict(plan_name="count")),
        (0, dict(scan_id={"$lt": 20}, plan_name="count")),
        (19, dict(scan_id={"$gte": 100})),
        (19, dict(scan_id={"$gte": 100}, plan_name="count")),
    ],
)
def test_ListRuns_query_parametrize(nruns, query, lr):
    lr.num = 100
    lr.query = query
    dd = lr.parse_runs()
    assert len(dd["time"]) == nruns
# fmt: on


# fmt: off
@pytest.mark.parametrize(
    "reverse",
    [
        (True),
        (False),
    ],
)
def test_ListRuns_reverse(reverse, lr):
    # include with some data missing or None
    lr.reverse = reverse
    dd = lr.parse_runs()
    assert dd["time"] == sorted(dd["time"], reverse=reverse)
# fmt: on


def test_ListRuns_since(lr):
    # depends on test database
    lr.num = 100
    dd = lr.parse_runs()  # get some data
    t = dd["time"]
    assert len(t) > 2
    lr.since = t[len(t) // 2]  # adjust the cutoff date
    dd = lr.parse_runs()  # get filtered data
    assert len(t) > len(dd["time"])
    for v in dd["time"]:
        assert v >= lr.since


def test_ListRuns_sortby(lr):
    # include with some data missing or None
    # default
    dd = lr.parse_runs()
    assert dd[lr.sortby] == sorted(dd[lr.sortby], reverse=lr.reverse)

    # other key: uid is almost always out of time order
    lr.sortby = "uid"
    lr.keys = "time uid".split()
    dd = lr.parse_runs()
    assert dd["time"] != sorted(dd["time"], reverse=lr.reverse)
    assert dd[lr.sortby] == sorted(dd[lr.sortby], reverse=lr.reverse)

    # these keys might not even be found
    lr.sortby = "motive"
    lr.keys += "motive purpose exit_status".split()
    dd = lr.parse_runs()
    assert dd[lr.sortby] == sorted(dd[lr.sortby], reverse=lr.reverse)


def test_ListRuns_timefmt(lr):
    # default
    dd = lr.parse_runs()
    assert "time" in dd
    assert len(dd["time"]) > 0
    v0 = dd["time"][0]
    assert isinstance(v0, str)
    dt = datetime.datetime.fromisoformat(v0)
    assert isinstance(dt, datetime.datetime)

    # None
    lr.timefmt = None
    with pytest.raises(TypeError) as exinfo:
        lr.parse_runs()
    assert "strftime() argument 1 must be str, not None" in str(exinfo.value)

    # wrong format
    lr.timefmt = "no such format"
    dd = lr.parse_runs()
    assert dd["time"][0] == lr.timefmt

    # "raw"
    lr.timefmt = "raw"
    dd = lr.parse_runs()
    assert isinstance(dd["time"][0], float)

    # "%X" : hh:mm:ss
    lr.timefmt = "%X"
    dd = lr.parse_runs()
    v0 = dd["time"][0]
    assert v0.find(":") == 2
    assert len(v0.split(":")) == 3


def test_ListRuns_to_dataframe(lr):
    # Pandas DataFrame
    lr.keys = None
    out = lr.to_dataframe()
    assert out is not None
    assert lr.keys is not None
    assert len(lr.keys) == 4
    assert isinstance(out, pd.DataFrame)
    assert len(out.columns) == len(lr.keys)
    for idx, label in enumerate(lr.keys):
        assert out.columns[idx] == label


def test_ListRuns_to_table(lr):
    # reST table as multiline string
    lr.keys = None
    out = lr.to_table()
    assert out is not None
    assert lr.keys is not None
    assert len(lr.keys) == 4
    assert isinstance(out, str)
    assert "=======" in out.splitlines()[0]
    assert "scan_id" in out.splitlines()[1]


def test_ListRuns_until(lr):
    # depends on test database
    lr.num = 100
    lr.reverse = False
    dd = lr.parse_runs()  # get some data
    t = dd["time"]
    assert len(t) > 2
    lr.until = t[len(t) // 2]  # adjust the cutoff date
    dd = lr.parse_runs()  # get filtered data
    assert len(t) > len(dd["time"])
    for v in dd["time"]:
        assert v < lr.until


# fmt: off
@pytest.mark.parametrize(
    "tablefmt, structure",
    [
        (None, pd.DataFrame),
        ("dataframe", pd.DataFrame),
        ("table", str),
        ("no such format", str),
    ],
)
def test_listruns_tablefmt(tablefmt, structure, cat):
    lr = utils.listruns(cat=cat, tablefmt=tablefmt, printing=False)
    assert lr is not None
    assert isinstance(lr, structure)
# fmt: on


# fmt: off
@pytest.mark.parametrize(
    "ids, nresults",
    [
        ([130, 131, 132], 3),
        ([1234], 0),
        (["bb7e0", ], 1),
        ([-2, -4, -10, -8], 4),
        ([], 0),
        ([-2, 131, "3e89a", 1234567, "1234567"], 3),
    ],
)
def test_ListRuns_ids(ids, nresults, lr):
    # include with some data missing or None
    lr.ids = ids
    dd = lr.parse_runs()
    assert len(dd["time"]) == nresults
# fmt: on



"""
Scaler support
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~use_EPICS_scaler_channels
"""

from ophyd import EpicsScaler
from ophyd.scaler import ScalerCH
import epics

# for convenience
SCALER_AUTOCOUNT_MODE = 1  # TODO: contribute to ophyd?


def use_EPICS_scaler_channels(scaler):
    """
    configure scaler for only the channels with names assigned in EPICS

    Note: For `ScalerCH`, use `scaler.select_channels(None)` instead of this code.
    (Applies only to `ophyd.scaler.ScalerCH` in releases after 2019-02-27.)
    """
    if isinstance(scaler, EpicsScaler):
        read_attrs = []
        for ch in scaler.channels.component_names:
            _nam = epics.caget(f"{scaler.prefix}.NM{int(ch[4:])}")
            if len(_nam.strip()) > 0:
                read_attrs.append(ch)
        scaler.channels.read_attrs = read_attrs
    elif isinstance(scaler, ScalerCH):
        # superceded by: https://github.com/NSLS-II/ophyd/commit/543e7ef81f3cb760192a0de719e51f9359642ae8
        scaler.match_names()
        read_attrs = []
        configuration_attrs = []
        for ch in scaler.channels.component_names:
            nm_pv = scaler.channels.__getattribute__(ch)
            if nm_pv is not None and len(nm_pv.chname.get().strip()) > 0:
                read_attrs.append(ch)
                configuration_attrs.append(ch)
                configuration_attrs.append(ch + ".chname")
                configuration_attrs.append(ch + ".preset")
                configuration_attrs.append(ch + ".gate")
        scaler.channels.read_attrs = read_attrs
        scaler.channels.configuration_attrs = configuration_attrs

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
db_2slit: synApps optics 2slit.db


There are two implementations, corresponding to differing and competing
opinions of how the support should be implemented.

Coordinates of ``Optics2Slit2D_InbOutBotTop`` (viewing from detector towards source)::

       top
    inb   out
       bot

Coordinates of ``Optics2Slit2D_HV`` (viewing from detector towards source)::

        v.xp
    h.xn    h.xp
        v.xn

Each blade [#]_ (in the XIA slit controller) travels in a _cartesian_ coordinate
system.  Positive motion moves a blade **outwards** (towards the ``p`` suffix).
Negative motion moves towards the ``n`` suffix.  Size and center are computed
by the underlying EPICS support.

    hsize = out - inb
    vsize = top - bot

..  [#] Note that the blade names here may be different than the EPICS support.
    The difference is to make the names of the blades consistent with other
    slits with the Bluesky framework.

USAGE::

    slit1 = Optics2Slit2D_HV("gp:Slit1", name="slit1")
    slit1.geometry = 0.1, 0.1, 0, 0  # moves the slits
    print(slit1.geometry)

    slit2 = Optics2Slit_InbOutBotTop("gp:Slit2", name="slit2")
    slit2.geometry = 0.1, 0.1, 0, 0  # moves the slits
    print(slit2.geometry)

Public Structures

.. autosummary::

    ~Optics2Slit1D
    ~Optics2Slit2D_HV
    ~Optics2Slit2D_InbOutBotTop

:see: https://github.com/epics-modules/optics

new in release 1.6.0
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal

from ..devices import PVPositionerSoftDone
from ..utils import SlitGeometry


class Optics2Slit1D(Device):
    """
    EPICS synApps optics 2slit.db 1D support: xn, xp, size, center, sync

    "sync" is used to tell the EPICS 2slit database to synchronize the
    virtual slit values with the actual motor positions.
    """
    xn = Component(
        PVPositionerSoftDone, "", setpoint_pv="xn", readback_pv="t2.B"
    )
    xp = Component(
        PVPositionerSoftDone, "", setpoint_pv="xp", readback_pv="t2.A"
    )
    size = Component(
        PVPositionerSoftDone, "", setpoint_pv="size", readback_pv="t2.C"
    )
    center = Component(
        PVPositionerSoftDone, "", setpoint_pv="center", readback_pv="t2.D"
    )

    sync = Component(EpicsSignal, "sync", put_complete=True, kind="omitted")


class Optics2Slit2D_HV(Device):
    """
    EPICS synApps optics 2slit.db 2D support: h.xn, h.xp, v.xn, v.xp
    """
    h = Component(Optics2Slit1D, "H")
    v = Component(Optics2Slit1D, "V")

    @property
    def geometry(self):
        """Return the slit 2D size and center as a namedtuple."""
        pppp = [
            round(obj.position, obj.precision)
            for obj in (
                self.h.size,
                self.v.size,
                self.h.center,
                self.v.center
            )
        ]

        return SlitGeometry(*pppp)

    @geometry.setter
    def geometry(self, value):
        # first, test the input by assigning it to local vars
        width, height, x, y = value

        self.h.size.move(width)
        self.v.size.move(height)
        self.h.center.move(x)
        self.v.center.move(y)


class Optics2Slit2D_InbOutBotTop(Device):
    """
    EPICS synApps optics 2slit.db 2D support: inb, out, bot, top
    """

    inb = Component(
        PVPositionerSoftDone, "H", setpoint_pv="xn", readback_pv="t2.B"
    )
    out = Component(
        PVPositionerSoftDone, "H", setpoint_pv="xp", readback_pv="t2.A"
    )
    hsize = Component(
        PVPositionerSoftDone, "H", setpoint_pv="size", readback_pv="t2.C"
    )
    hcenter = Component(
        PVPositionerSoftDone, "H", setpoint_pv="center", readback_pv="t2.D"
    )

    bot = Component(
        PVPositionerSoftDone, "V", setpoint_pv="xn", readback_pv="t2.B"
    )
    top = Component(
        PVPositionerSoftDone, "V", setpoint_pv="xp", readback_pv="t2.A"
    )
    vsize = Component(
        PVPositionerSoftDone, "V", setpoint_pv="size", readback_pv="t2.C"
    )
    vcenter = Component(
        PVPositionerSoftDone, "V", setpoint_pv="center", readback_pv="t2.D"
    )

    hsync = Component(EpicsSignal, "Hsync", put_complete=True, kind="omitted")
    vsync = Component(EpicsSignal, "Vsync", put_complete=True, kind="omitted")

    @property
    def geometry(self):
        """Return the slit 2D size and center as a namedtuple."""
        pppp = [
            round(obj.position, obj.precision)
            for obj in (
                self.hsize,
                self.vsize,
                self.hcenter,
                self.vcenter
            )
        ]

        return SlitGeometry(*pppp)

    @geometry.setter
    def geometry(self, value):
        # first, test the input by assigning it to local vars
        width, height, x, y = value

        self.hsize.move(width)
        self.vsize.move(height)
        self.hcenter.move(x)
        self.vcenter.move(y)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
